# coding=utf-8
# Copyright (c) 2021, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""GPT-2 model."""
import enum
import math
import contextlib
import json

import torch
import torch.nn.functional as F

import apex.transformer.utils
from apex.normalization import FusedLayerNorm as LayerNorm
from apex.transformer.functional import FusedScaleMaskSoftmax
from apex.transformer import tensor_parallel
from apex.transformer import parallel_state
from apex.transformer.testing.global_vars import get_args
from apex.transformer.enums import LayerType
from apex.transformer.enums import AttnType
from apex.transformer.enums import AttnMaskType


_FLOAT_TYPES = (torch.FloatTensor, torch.cuda.FloatTensor)
_HALF_TYPES = (torch.HalfTensor, torch.cuda.HalfTensor)
_BF16_TYPES = (torch.BFloat16Tensor, torch.cuda.BFloat16Tensor)


class ModelType(enum.Enum):
    encoder_or_decoder = 1
    encoder_and_decoder = 2


###### BIAS GELU FUSION/ NO AUTOGRAD ################
# 1/sqrt(2*pi)-> 0.3989423
# 1/sqrt(2)   -> 0.70710678
# sqrt(2/pi)  -> 0.79788456
# this function is tanh approximation of gelu
# actual gelu is:
# x * 0.5 * (1.0 + torch.erf(x * 0.70710678))


@torch.jit.script
def bias_gelu(bias, y):
    x = bias + y
    return x * 0.5 * (1.0 + torch.tanh(0.79788456 * x * (1 + 0.044715 * x * x)))


# gradient of tanh approximation of gelu
# gradient of actual gelu is:
# 0.5 * (1. + torch.erf(x * 0.70710678)) + 0.3989423 * x * torch.exp(-0.5 * x * x)
@torch.jit.script
def bias_gelu_back(g, bias, y):
    x = bias + y
    tanh_out = torch.tanh(0.79788456 * x * (1 + 0.044715 * x * x))
    # sqrt(2/pi) * 3 * 0.044715 -> 0.1070322243
    ff = 0.5 * x * ((1 - tanh_out * tanh_out) * (0.79788456 + 0.1070322243 * x * x)) + 0.5 * (1 + tanh_out)
    return ff * g

class MegatronModule(torch.nn.Module):
    """Megatron specific extensions of torch Module with support
    for pipelining."""

    def __init__(self, share_word_embeddings=True):
        super(MegatronModule, self).__init__()
        self.share_word_embeddings = share_word_embeddings


    def state_dict_for_save_checkpoint(self, destination=None, prefix="",
                                       keep_vars=False):
        """Use this function to override the state dict for
        saving checkpoints."""
        return self.state_dict(destination, prefix, keep_vars)


    def word_embeddings_weight(self):
        if not parallel_state.is_pipeline_last_stage(ignore_virtual=True) or \
                parallel_state.get_pipeline_model_parallel_world_size() == 1:
            return self.language_model.embedding.word_embeddings.weight
        else:
            if not self.share_word_embeddings:
                raise Exception("word_embeddings_weight() called for last "
                                "stage, but share_word_embeddings is false")
            return self.word_embeddings.weight


    def initialize_word_embeddings(self, init_method_normal):
        args = get_args()
        if not self.share_word_embeddings:
            raise Exception("initialize_word_embeddings() was called but "
                            "share_word_embeddings is false")

        # This function just initializes the word embeddings in the final stage
        # when we are using pipeline parallelism. Nothing to do if we aren't
        # using pipeline parallelism.
        if args.pipeline_model_parallel_size == 1:
            return

        # Parameters are shared between the word embeddings layers, and the
        # heads at the end of the model. In a pipelined setup with more than
        # one stage, the initial embedding layer and the head are on different
        # workers, so we do the following:
        # 1. Create a second copy of word_embeddings on the last stage, with
        #    initial parameters of 0.0.
        # 2. Do an all-reduce between the first and last stage to ensure that
        #    the two copies of word_embeddings start off with the same
        #    parameter values.
        # 3. In the training loop, before an all-reduce between the grads of
        #    the two word_embeddings layers to ensure that every applied weight
        #    update is the same on both stages.
        if parallel_state.is_pipeline_last_stage():
            assert not parallel_state.is_pipeline_first_stage()
            self._word_embeddings_for_head_key = "word_embeddings_for_head"
            # set word_embeddings weights to 0 here, then copy first
            # stage's weights using all_reduce below.
            self.word_embeddings = tensor_parallel.VocabParallelEmbedding(
                args.padded_vocab_size, args.hidden_size,
                init_method=init_method_normal(args.init_method_std),
		use_cpu_initialization=args.use_cpu_initialization)
            self.word_embeddings.weight.data.fill_(0)
            self.word_embeddings.weight.shared = True

        # Zero out initial weights for decoder embedding.
        # NOTE: We don't currently support T5 with the interleaved schedule.
        if not parallel_state.is_pipeline_first_stage(ignore_virtual=True) and \
                not parallel_state.is_pipeline_last_stage(ignore_virtual=True) and \
                parallel_state.is_rank_in_embedding_group():
            self.language_model.embedding.zero_parameters()

        # Ensure that first and last stages have the same initial parameter
        # values.
        if torch.distributed.is_initialized():
            if parallel_state.is_rank_in_embedding_group():
                torch.distributed.all_reduce(self.word_embeddings_weight().data,
                                             group=parallel_state.get_embedding_group())
                # All-reduce other embeddings as well as necessary. The last stage
                # does not have these other embeddings, so just create placeholder
                # tensors of the right shape with all zeros.
                # NOTE: We don't currently support T5 with the interleaved schedule.
                if args.pipeline_model_parallel_split_rank is not None:
                    # TODO: Support tokentype embedding.
                    dimensions = (args.max_position_embeddings, args.hidden_size)
                    if parallel_state.is_pipeline_last_stage(ignore_virtual=True):
                        position_embeddings = torch.nn.Embedding(*dimensions).cuda()
                        position_embeddings.weight.data.fill_(0)
                    else:
                        self.language_model.embedding.cuda()
                        position_embeddings = self.language_model.embedding.position_embeddings
                    torch.distributed.all_reduce(position_embeddings.weight.data,
                                                 group=parallel_state.get_embedding_group())
        else:
            print("WARNING! Distributed processes aren't initialized, so "
                  "word embeddings in the last layer are not initialized. "
                  "If you are just manipulating a model this is fine, but "
                  "this needs to be handled manually. If you are training "
                  "something is definitely wrong.")

class GeLUFunction(torch.autograd.Function):
    @staticmethod
    # bias is an optional argument
    def forward(ctx, input, bias):
        ctx.save_for_backward(input, bias)
        return bias_gelu(bias, input)

    @staticmethod
    def backward(ctx, grad_output):
        input, bias = ctx.saved_tensors
        tmp = bias_gelu_back(grad_output, bias, input)
        return tmp, tmp


bias_gelu_impl = GeLUFunction.apply


def get_linear_layer(rows, columns, init_method):
    """Simple linear layer with weight initialization."""
    layer = torch.nn.Linear(rows, columns)
    init_method(layer.weight)
    with torch.no_grad():
        layer.bias.zero_()
    return layer


def attention_mask_func(attention_scores, attention_mask):
    attention_scores.masked_fill_(attention_mask, -10000.0)
    return attention_scores


@torch.jit.script
def gelu_impl(x):
    """OpenAI's gelu implementation."""
    return 0.5 * x * (1.0 + torch.tanh(0.7978845608028654 * x * (1.0 + 0.044715 * x * x)))


def openai_gelu(x):
    return gelu_impl(x)


# This is actually Python equivalent of torch.nn.functional.gelu(), also with type hints for ONNX exporter
@torch.jit.script
def erf_gelu(x):
    return x * 0.5 * (torch.erf(x / 1.41421).to(dtype=x.dtype) + torch.ones_like(x).to(dtype=x.dtype))


def init_method_normal(sigma):
    """Init method based on N(0, sigma)."""

    def init_(tensor):
        return torch.nn.init.normal_(tensor, mean=0.0, std=sigma)

    return init_


def scaled_init_method_normal(sigma, num_layers):
    """Init method based on N(0, sigma/sqrt(2*num_layers)."""
    std = sigma / math.sqrt(2.0 * num_layers)

    def init_(tensor):
        return torch.nn.init.normal_(tensor, mean=0.0, std=std)

    return init_


class ParallelMLP(MegatronModule):
    """MLP.

    MLP will take the input with h hidden state, project it to 4*h
    hidden dimension, perform nonlinear transformation, and project the
    state back into h hidden dimension.
    """

    def __init__(self, init_method, output_layer_init_method):
        super().__init__()
        args = get_args()

        # Project to 4h.
        self.dense_h_to_4h = tensor_parallel.ColumnParallelLinear(
            args.hidden_size, args.ffn_hidden_size, gather_output=False, init_method=init_method, skip_bias_add=True,
        use_cpu_initialization=args.use_cpu_initialization)

        self.bias_gelu_fusion = args.bias_gelu_fusion
        self.activation_func = F.gelu
        if args.openai_gelu:
            self.activation_func = openai_gelu
        elif args.onnx_safe:
            self.activation_func = erf_gelu

        # Project back to h.
        self.dense_4h_to_h = tensor_parallel.RowParallelLinear(
            args.ffn_hidden_size,
            args.hidden_size,
            input_is_parallel=True,
            init_method=output_layer_init_method,
            skip_bias_add=True,
            use_cpu_initialization=args.use_cpu_initialization
        )

    def forward(self, hidden_states):

        # [s, b, 4hp]
        intermediate_parallel, bias_parallel = self.dense_h_to_4h(hidden_states)

        if self.bias_gelu_fusion:
            intermediate_parallel = bias_gelu_impl(intermediate_parallel, bias_parallel)
        else:
            intermediate_parallel = self.activation_func(intermediate_parallel + bias_parallel)

        # [s, b, h]
        output, output_bias = self.dense_4h_to_h(intermediate_parallel)
        return output, output_bias


class ParallelAttention(MegatronModule):
    """Parallel self-attention layer abstract class.

    Self-attention layer takes input with size [b, s, h]
    and returns output of the same size.
    """

    def __init__(
        self,
        init_method,
        output_layer_init_method,
        layer_number,
        attention_type=AttnType.self_attn,
        attn_mask_type=AttnMaskType.padding,
    ):
        super().__init__()
        args = get_args()
        self.fp16 = args.fp16
        self.bf16 = args.bf16

        self.apply_query_key_layer_scaling = args.apply_query_key_layer_scaling
        self.attention_softmax_in_fp32 = args.attention_softmax_in_fp32
        if self.apply_query_key_layer_scaling:
            self.attention_softmax_in_fp32 = True
        self.layer_number = max(1, layer_number)
        self.attention_type = attention_type
        self.attn_mask_type = attn_mask_type
        self.params_dtype = args.params_dtype

        projection_size = args.kv_channels * args.num_attention_heads

        # Per attention head and per partition values.
        world_size = parallel_state.get_tensor_model_parallel_world_size()
        self.hidden_size_per_partition = apex.transformer.utils.divide(projection_size, world_size)
        self.hidden_size_per_attention_head = apex.transformer.utils.divide(projection_size, args.num_attention_heads)
        self.num_attention_heads_per_partition = apex.transformer.utils.divide(args.num_attention_heads, world_size)

        # Strided linear layer.
        if attention_type == AttnType.self_attn:
            self.query_key_value = tensor_parallel.ColumnParallelLinear(
                args.hidden_size, 3 * projection_size, gather_output=False, init_method=init_method, use_cpu_initialization=args.use_cpu_initialization)
        else:
            assert attention_type == AttnType.cross_attn
            self.query = tensor_parallel.ColumnParallelLinear(
                args.hidden_size, projection_size, gather_output=False, init_method=init_method, use_cpu_initialization=args.use_cpu_initialization)

            self.key_value = tensor_parallel.ColumnParallelLinear(
                args.hidden_size, 2 * projection_size, gather_output=False, init_method=init_method, use_cpu_initialization=args.use_cpu_initialization)

        coeff = None
        self.norm_factor = math.sqrt(self.hidden_size_per_attention_head)
        if self.apply_query_key_layer_scaling:
            coeff = self.layer_number
            self.norm_factor *= coeff

        self.scale_mask_softmax = FusedScaleMaskSoftmax(
            self.fp16,
            self.bf16,
            self.attn_mask_type,
            args.masked_softmax_fusion,
            attention_mask_func,
            self.attention_softmax_in_fp32,
            coeff,
        )

        # Dropout. Note that for a single iteration, this layer will generate
        # different outputs on different number of parallel partitions but
        # on average it should not be partition dependent.
        self.attention_dropout = torch.nn.Dropout(args.attention_dropout)

        # Output.
        self.dense = tensor_parallel.RowParallelLinear(
            projection_size,
            args.hidden_size,
            input_is_parallel=True,
            init_method=output_layer_init_method,
            skip_bias_add=True,
            use_cpu_initialization=args.use_cpu_initialization
        )

        # Inference key-value memory
        self.inference_key_memory = None
        self.inference_value_memory = None
        self.inference_current_sequence_len = 0

    def _allocate_memory(self, inference_max_sequence_len, batch_size):
        return torch.empty(
            inference_max_sequence_len,
            batch_size,
            self.num_attention_heads_per_partition,
            self.hidden_size_per_attention_head,
            dtype=self.params_dtype,
            device=torch.cuda.current_device(),
        )

    def forward(
        self,
        hidden_states,
        attention_mask,
        encoder_output=None,
        set_inference_key_value_memory=False,
        inference_max_sequence_len=None,
    ):
        # hidden_states: [sq, b, h]

        # =================================================
        # Pre-allocate memory for key-values for inference.
        # =================================================
        if set_inference_key_value_memory:
            assert inference_max_sequence_len and inference_max_sequence_len > 0
            self.inference_key_memory = self._allocate_memory(inference_max_sequence_len, hidden_states.size(1))
            self.inference_value_memory = self._allocate_memory(inference_max_sequence_len, hidden_states.size(1))
            self.inference_current_sequence_len = 0
        # Some consistency check.
        if inference_max_sequence_len:
            assert self.inference_current_sequence_len < self.inference_key_memory.size(0)
            assert inference_max_sequence_len == self.inference_key_memory.size(0)
        # This is added for safety. In case inference_max_sequence_len
        # is not provided, make sure there is no potential memory left
        # from previous inference.
        if not inference_max_sequence_len:
            self.inference_key_memory = None
            self.inference_value_memory = None

        # =====================
        # Query, Key, and Value
        # =====================

        if self.attention_type == AttnType.self_attn:
            # Attention heads [sq, b, h] --> [sq, b, (np * 3 * hn)]
            mixed_x_layer, _ = self.query_key_value(hidden_states)

            # [sq, b, (np * 3 * hn)] --> [sq, b, np, 3 * hn]
            new_tensor_shape = mixed_x_layer.size()[:-1] + (
                self.num_attention_heads_per_partition,
                3 * self.hidden_size_per_attention_head,
            )
            mixed_x_layer = mixed_x_layer.view(*new_tensor_shape)

            # [sq, b, np, 3 * hn] --> 3 [sq, b, np, hn]
            (query_layer, key_layer, value_layer) = tensor_parallel.split_tensor_along_last_dim(mixed_x_layer, 3)
        else:
            # Attention heads [sk, b, h] --> [sk, b, (np * 2 * hn)]
            mixed_kv_layer, _ = self.key_value(encoder_output)

            # [sk, b, (np * 2 * hn)] --> [sk, b, np, 2 * hn]
            new_tensor_shape = mixed_kv_layer.size()[:-1] + (
                self.num_attention_heads_per_partition,
                2 * self.hidden_size_per_attention_head,
            )
            mixed_kv_layer = mixed_kv_layer.view(*new_tensor_shape)

            # [sk, b, np, 2 * hn] --> 2 [sk, b, np, hn]
            (key_layer, value_layer) = tensor_parallel.split_tensor_along_last_dim(mixed_kv_layer, 2)

            # Attention head [sq, b, h] --> [sq, b, hp]
            query_layer, _ = self.query(hidden_states)
            # [sq, b, hp] --> [sq, b, np, hn]
            new_tensor_shape = query_layer.size()[:-1] + (
                self.num_attention_heads_per_partition,
                self.hidden_size_per_attention_head,
            )
            query_layer = query_layer.view(*new_tensor_shape)

        # ===================================================
        # Adjust key, value, and attention mask for inference
        # ===================================================

        if inference_max_sequence_len:
            # Adjust the range variables.
            start = self.inference_current_sequence_len
            self.inference_current_sequence_len += key_layer.size(0)
            end = self.inference_current_sequence_len
            # Copy key and values.
            self.inference_key_memory[start:end, ...] = key_layer
            self.inference_value_memory[start:end, ...] = value_layer
            key_layer = self.inference_key_memory[:end, ...]
            value_layer = self.inference_value_memory[:end, ...]
            # Adjust attention mask
            attention_mask = attention_mask[..., start:end, :end]

        # ===================================
        # Raw attention scores. [b, np, s, s]
        # ===================================

        # [b, np, sq, sk]
        output_size = (query_layer.size(1), query_layer.size(2), query_layer.size(0), key_layer.size(0))

        # [sq, b, np, hn] -> [sq, b * np, hn]
        query_layer = query_layer.view(output_size[2], output_size[0] * output_size[1], -1)
        # [sk, b, np, hn] -> [sk, b * np, hn]
        key_layer = key_layer.view(output_size[3], output_size[0] * output_size[1], -1)

        # preallocting result tensor: [b * np, sq, sk]
        matmul_result = torch.empty(
            output_size[0] * output_size[1],
            output_size[2],
            output_size[3],
            dtype=query_layer.dtype,
            device=torch.cuda.current_device(),
        )

        # Raw attention scores. [b * np, sq, sk]
        matmul_result = torch.baddbmm(
            matmul_result,
            query_layer.transpose(0, 1),  # [b * np, sq, hn]
            key_layer.transpose(0, 1).transpose(1, 2),  # [b * np, hn, sk]
            beta=0.0,
            alpha=(1.0 / self.norm_factor),
        )

        # change view to [b, np, sq, sk]
        attention_scores = matmul_result.view(*output_size)

        # ===========================
        # Attention probs and dropout
        # ===========================

        # attention scores and attention mask [b, np, sq, sk]
        attention_probs = self.scale_mask_softmax(attention_scores, attention_mask)

        # This is actually dropping out entire tokens to attend to, which might
        # seem a bit unusual, but is taken from the original Transformer paper.
        with tensor_parallel.get_cuda_rng_tracker().fork():
            attention_probs = self.attention_dropout(attention_probs)

        # =========================
        # Context layer. [sq, b, hp]
        # =========================

        # value_layer -> context layer.
        # [sk, b, np, hn] --> [b, np, sq, hn]

        # context layer shape: [b, np, sq, hn]
        output_size = (value_layer.size(1), value_layer.size(2), query_layer.size(0), value_layer.size(3))

        # change view [sk, b * np, hn]
        value_layer = value_layer.view(value_layer.size(0), output_size[0] * output_size[1], -1)

        # change view [b * np, sq, sk]
        attention_probs = attention_probs.view(output_size[0] * output_size[1], output_size[2], -1)

        # matmul: [b * np, sq, hn]
        context_layer = torch.bmm(attention_probs, value_layer.transpose(0, 1))

        # change view [b, np, sq, hn]
        context_layer = context_layer.view(*output_size)

        # [b, np, sq, hn] --> [sq, b, np, hn]
        context_layer = context_layer.permute(2, 0, 1, 3).contiguous()

        # [sq, b, np, hn] --> [sq, b, hp]
        new_context_layer_shape = context_layer.size()[:-2] + (self.hidden_size_per_partition,)
        context_layer = context_layer.view(*new_context_layer_shape)

        # =================
        # Output. [sq, b, h]
        # =================

        output, bias = self.dense(context_layer)

        return output, bias


@torch.jit.script
def bias_dropout_add(x: torch.Tensor, bias: torch.Tensor, residual: torch.Tensor, prob: float, training: bool) -> torch.Tensor:
    out = torch.nn.functional.dropout(x + bias, p=prob, training=training)
    out = residual + out
    return out


def get_bias_dropout_add(training):
    def _bias_dropout_add(x, bias, residual, prob):
        return bias_dropout_add(x, bias, residual, prob, training)

    return _bias_dropout_add


@torch.jit.script
def bias_dropout_add_fused_train(
    x: torch.Tensor, bias: torch.Tensor, residual: torch.Tensor, prob: float
) -> torch.Tensor:
    return bias_dropout_add(x, bias, residual, prob, True)


@torch.jit.script
def bias_dropout_add_fused_inference(
    x: torch.Tensor, bias: torch.Tensor, residual: torch.Tensor, prob: float
) -> torch.Tensor:
    return bias_dropout_add(x, bias, residual, prob, False)


class ParallelTransformerLayer(MegatronModule):
    """A single transformer layer.

    Transformer layer takes input with size [b, s, h] and returns an
    output of the same size.
    """

    def __init__(
        self,
        init_method,
        output_layer_init_method,
        layer_number,
        layer_type=LayerType.encoder,
        self_attn_mask_type=AttnMaskType.padding,
    ):
        args = get_args()

        super().__init__()
        self.layer_number = layer_number
        self.layer_type = layer_type

        self.apply_residual_connection_post_layernorm = args.apply_residual_connection_post_layernorm

        self.bf16 = args.bf16
        self.fp32_residual_connection = args.fp32_residual_connection

        # Layernorm on the input data.
        self.input_layernorm = LayerNorm(args.hidden_size, eps=args.layernorm_epsilon)

        # Self attention.
        self.self_attention = ParallelAttention(
            init_method,
            output_layer_init_method,
            layer_number,
            attention_type=AttnType.self_attn,
            attn_mask_type=self_attn_mask_type,
        )
        self.hidden_dropout = args.hidden_dropout
        self.bias_dropout_fusion = args.bias_dropout_fusion

        # Layernorm on the attention output
        self.post_attention_layernorm = LayerNorm(args.hidden_size, eps=args.layernorm_epsilon)

        if self.layer_type == LayerType.decoder:
            self.inter_attention = ParallelAttention(
                init_method, output_layer_init_method, layer_number, attention_type=AttnType.cross_attn
            )
            # Layernorm on the attention output.
            self.post_inter_attention_layernorm = LayerNorm(args.hidden_size, eps=args.layernorm_epsilon)

        # MLP
        self.mlp = ParallelMLP(init_method, output_layer_init_method)

    def forward(
        self,
        hidden_states,
        attention_mask,
        encoder_output=None,
        enc_dec_attn_mask=None,
        set_inference_key_value_memory=False,
        inference_max_sequence_len=None,
    ):
        # hidden_states: [b, s, h]

        # Layer norm at the beginning of the transformer layer.
        layernorm_output = self.input_layernorm(hidden_states)
        # Self attention.
        attention_output, attention_bias = self.self_attention(
            layernorm_output,
            attention_mask,
            set_inference_key_value_memory=set_inference_key_value_memory,
            inference_max_sequence_len=inference_max_sequence_len,
        )

        # Residual connection.
        if self.apply_residual_connection_post_layernorm:
            residual = layernorm_output
        else:
            residual = hidden_states

        # jit scripting for a nn.module (with dropout) is not
        # trigerring the fusion kernel. For now, we use two
        # different nn.functional routines to account for varying
        # dropout semantics during training and inference phases.
        if self.bias_dropout_fusion:
            if self.training:
                bias_dropout_add_func = bias_dropout_add_fused_train
            else:
                bias_dropout_add_func = bias_dropout_add_fused_inference
        else:
            bias_dropout_add_func = get_bias_dropout_add(self.training)

        # re-enable torch grad to enable fused optimization.
        with torch.enable_grad():
            layernorm_input = bias_dropout_add_func(
                attention_output, attention_bias.expand_as(residual), residual, self.hidden_dropout
            )

        # Layer norm post the self attention.
        layernorm_output = self.post_attention_layernorm(layernorm_input)

        if self.layer_type == LayerType.decoder:
            attention_output, attention_bias = self.inter_attention(
                layernorm_output, enc_dec_attn_mask, encoder_output=encoder_output
            )
            # residual connection
            if self.apply_residual_connection_post_layernorm:
                residual = layernorm_output
            else:
                residual = layernorm_input

            # re-enable torch grad to enable fused optimization.
            with torch.enable_grad():
                layernorm_input = bias_dropout_add_func(
                    attention_output, attention_bias.expand_as(residual), residual, self.hidden_dropout
                )

            # Layer norm post the decoder attention
            layernorm_output = self.post_inter_attention_layernorm(layernorm_input)

        # MLP.
        mlp_output, mlp_bias = self.mlp(layernorm_output)

        # Second residual connection.
        if self.apply_residual_connection_post_layernorm:
            residual = layernorm_output
        else:
            residual = layernorm_input

        # re-enable torch grad to enable fused optimization.
        with torch.enable_grad():
            output = bias_dropout_add_func(mlp_output, mlp_bias.expand_as(residual), residual, self.hidden_dropout)

        return output


class ParallelTransformer(MegatronModule):
    """Transformer class."""

    def __init__(
        self,
        init_method,
        output_layer_init_method,
        layer_type=LayerType.encoder,
        self_attn_mask_type=AttnMaskType.padding,
        pre_process=True,
        post_process=True,
    ):
        super().__init__()
        args = get_args()

        self.bf16 = args.bf16
        self.fp32_residual_connection = args.fp32_residual_connection
        self.pre_process = pre_process
        self.post_process = post_process
        self.input_tensor = None

        # Store activation checkpointing flag.
        self.activations_checkpoint_method = args.activations_checkpoint_method
        self.activations_checkpoint_num_layers = args.activations_checkpoint_num_layers
        self.distribute_checkpointed_activations = args.distribute_checkpointed_activations

        num_layers = args.num_layers
        # Number of layers.
        assert (
            num_layers % parallel_state.get_pipeline_model_parallel_world_size() == 0
        ), "num_layers must be divisible by pipeline_model_parallel_size"
        self.num_layers = num_layers // parallel_state.get_pipeline_model_parallel_world_size()

        # Transformer layers.
        def build_layer(layer_number):
            return ParallelTransformerLayer(
                init_method,
                output_layer_init_method,
                layer_number,
                layer_type=layer_type,
                self_attn_mask_type=self_attn_mask_type,
            )

        if args.virtual_pipeline_model_parallel_size is not None:
            assert args.num_layers % args.virtual_pipeline_model_parallel_size == 0, (
                "num_layers_per_stage must be divisible by " "virtual_pipeline_model_parallel_size"
            )
            # Number of layers in each model chunk is the number of layers in the stage,
            # divided by the number of model chunks in a stage.
            self.num_layers = self.num_layers // args.virtual_pipeline_model_parallel_size
            # With 8 layers, 2 stages, and 4 model chunks, we want an assignment of
            # layers to stages like (each list is a model chunk):
            # Stage 0: [0]  [2]  [4]  [6]
            # Stage 1: [1]  [3]  [5]  [7]
            # With 8 layers, 2 stages, and 2 virtual stages, we want an assignment of
            # layers to stages like (each list is a model chunk):
            # Stage 0: [0, 1]  [4, 5]
            # Stage 1: [2, 3]  [6, 7]
            offset = parallel_state.get_virtual_pipeline_model_parallel_rank() * (
                args.num_layers // args.virtual_pipeline_model_parallel_size
            ) + (parallel_state.get_pipeline_model_parallel_rank() * self.num_layers)
        else:
            # Each stage gets a contiguous set of layers.
            offset = parallel_state.get_pipeline_model_parallel_rank() * self.num_layers

        self.layers = torch.nn.ModuleList([build_layer(i + 1 + offset) for i in range(self.num_layers)])

        if self.post_process:
            # Final layer norm before output.
            self.final_layernorm = LayerNorm(args.hidden_size, eps=args.layernorm_epsilon)

    def _get_layer(self, layer_number):
        return self.layers[layer_number]

    def _checkpointed_forward(self, hidden_states, attention_mask, encoder_output, enc_dec_attn_mask):
        """Forward method with activation checkpointing."""

        def custom(start, end):
            def custom_forward(*inputs):
                x_ = inputs[0]
                attention_mask = inputs[1]
                encoder_output = inputs[2]
                enc_dec_attn_mask = inputs[3]
                for index in range(start, end):
                    layer = self._get_layer(index)
                    x_ = layer(x_, attention_mask, encoder_output, enc_dec_attn_mask)
                return x_

            return custom_forward

        def distribute_checkpointed_activations_helper(layer_number):
            """Distribute checkpointed activations across the tensor model
               Parallel ranks if the `distribute-checkpointed-activations
               is on and either of the following conditions is met:
                 - it is not the first layer in the in the pipeline stage.
                   The first layer is used in the pipeline parallelism
                   and changing its shape throws error in the backward pass.
                 - we are at the first pipline stage so the input tensor is
                   not used in pipeline parallelism. Note that no pipeline
                   parallelism is a special case of this.
            """
            not_first_layer_in_pipeline_stage = layer_number > 0
            is_first_pipeline_stage = parallel_state.get_pipeline_model_parallel_rank() == 0
            return self.distribute_checkpointed_activations and (
                not_first_layer_in_pipeline_stage or is_first_pipeline_stage
            )

        if self.activations_checkpoint_method == "uniform":
            # Uniformly divide the total number of Transformer layers and checkpoint
            # the input activation of each divided chunk.
            # A method to further reduce memory usage reducing checkpoints.
            l = 0
            while l < self.num_layers:
                hidden_states = tensor_parallel.checkpoint(
                    custom(l, l + self.activations_checkpoint_num_layers),
                    distribute_checkpointed_activations_helper(l),
                    hidden_states,
                    attention_mask,
                    encoder_output,
                    enc_dec_attn_mask,
                )
                l += self.activations_checkpoint_num_layers
        elif self.activations_checkpoint_method == "block":
            # Checkpoint the input activation of only a set number of individual
            # Transformer layers and skip the rest.
            # A method fully use the device memory removing redundant re-computation.
            for l in range(self.num_layers):
                if l < self.activations_checkpoint_num_layers:
                    hidden_states = tensor_parallel.checkpoint(
                        custom(l, l + 1),
                        distribute_checkpointed_activations_helper(l),
                        hidden_states,
                        attention_mask,
                        encoder_output,
                        enc_dec_attn_mask,
                    )
                else:
                    hidden_states = custom(l, l + 1)(hidden_states, attention_mask, encoder_output, enc_dec_attn_mask)
        else:
            raise ValueError("Invalid activation checkpoint method.")

        return hidden_states

    def set_input_tensor(self, input_tensor):
        """Set input tensor to be used instead of forward()'s input.

        When doing pipeline parallelism the input from the previous
        stage comes from communication, not from the input, so the
        model's forward_step_func won't have it. This function is thus
        used by internal code to bypass the input provided by the
        forward_step_func"""
        self.input_tensor = input_tensor

    def forward(
        self,
        hidden_states,
        attention_mask,
        encoder_output=None,
        enc_dec_attn_mask=None,
        set_inference_key_value_memory=False,
        inference_max_sequence_len=None,
    ):

        # Checks.
        if inference_max_sequence_len:
            assert self.activations_checkpoint_method is None, "inference does not work with activation checkpointing"

        if self.pre_process:
            # Data format change to avoid explicit tranposes : [b s h] --> [s b h].
            # If the input flag for fp32 residual connection is set, convert for float.
            if self.fp32_residual_connection:
                hidden_states = hidden_states.transpose(0, 1).contiguous().float()
            # Otherwise, leave it as is.
            else:
                hidden_states = hidden_states.transpose(0, 1).contiguous()
        else:
            # See set_input_tensor()
            hidden_states = self.input_tensor

        if encoder_output is not None:
            encoder_output = encoder_output.transpose(0, 1).contiguous()

        if self.activations_checkpoint_method is not None:
            hidden_states = self._checkpointed_forward(hidden_states, attention_mask, encoder_output, enc_dec_attn_mask)
        else:
            for index in range(self.num_layers):
                layer = self._get_layer(index)
                hidden_states = layer(
                    hidden_states,
                    attention_mask,
                    encoder_output=encoder_output,
                    enc_dec_attn_mask=enc_dec_attn_mask,
                    set_inference_key_value_memory=set_inference_key_value_memory,
                    inference_max_sequence_len=inference_max_sequence_len,
                )

        # Final layer norm.
        if self.post_process:
            # Reverting data format change [s b h] --> [b s h].
            hidden_states = hidden_states.transpose(0, 1).contiguous()
            output = self.final_layernorm(hidden_states)
        else:
            output = hidden_states

        return output


def parallel_lm_logits(input_, word_embeddings_weight, parallel_output, bias=None):
    """LM logits using word embedding weights."""
    # Parallel logits.
    input_parallel = tensor_parallel.copy_to_tensor_model_parallel_region(input_)
    # Matrix multiply.
    if bias is None:
        logits_parallel = F.linear(input_parallel, word_embeddings_weight)
    else:
        logits_parallel = F.linear(input_parallel, word_embeddings_weight, bias)
    # Gather if needed.
    if parallel_output:
        return logits_parallel

    return tensor_parallel.gather_from_tensor_model_parallel_region(logits_parallel)


def get_language_model(
    num_tokentypes,
    add_pooler,
    encoder_attn_mask_type,
    init_method=None,
    scaled_init_method=None,
    add_encoder=True,
    add_decoder=False,
    decoder_attn_mask_type=AttnMaskType.causal,
    pre_process=True,
    post_process=True,
):
    """Build language model and return along with the key to save."""
    args = get_args()

    if init_method is None:
        init_method = init_method_normal(args.init_method_std)
    if scaled_init_method is None:
        scaled_init_method = scaled_init_method_normal(args.init_method_std, args.num_layers)

    # Language model.
    language_model = TransformerLanguageModel(
        init_method,
        scaled_init_method,
        encoder_attn_mask_type,
        num_tokentypes=num_tokentypes,
        add_encoder=add_encoder,
        add_decoder=add_decoder,
        decoder_attn_mask_type=decoder_attn_mask_type,
        add_pooler=add_pooler,
        pre_process=pre_process,
        post_process=post_process,
    )
    # key used for checkpoints.
    language_model_key = "language_model"

    return language_model, language_model_key


class Pooler(MegatronModule):
    """Pooler layer.

    Pool hidden states of a specific token (for example start of the
    sequence) and add a linear transformation followed by a tanh.

    Arguments:
        hidden_size: hidden size
        init_method: weight initialization method for the linear layer.
            bias is set to zero.
    """

    def __init__(self, hidden_size, init_method):
        super(Pooler, self).__init__()
        self.dense = get_linear_layer(hidden_size, hidden_size, init_method)

    def forward(self, hidden_states, sequence_index=0):
        # hidden_states: [b, s, h]
        # sequence_index: index of the token to pool.
        pooled = hidden_states[:, sequence_index, :]
        pooled = self.dense(pooled)
        pooled = torch.tanh(pooled)
        return pooled


class Embedding(MegatronModule):
    """Language model embeddings.

    Arguments:
        hidden_size: hidden size
        vocab_size: vocabulary size
        max_sequence_length: maximum size of sequence. This
                             is used for positional embedding
        embedding_dropout_prob: dropout probability for embeddings
        init_method: weight initialization method
        num_tokentypes: size of the token-type embeddings. 0 value
                        will ignore this embedding
    """

    def __init__(
        self, hidden_size, vocab_size, max_sequence_length, embedding_dropout_prob, init_method, num_tokentypes=0
    ):
        super().__init__()

        self.hidden_size = hidden_size
        self.init_method = init_method
        self.num_tokentypes = num_tokentypes
        args = get_args()

        # Word embeddings (parallel).
        self.word_embeddings = tensor_parallel.VocabParallelEmbedding(
            vocab_size, self.hidden_size, init_method=self.init_method,
            use_cpu_initialization=args.use_cpu_initialization
        )
        self._word_embeddings_key = "word_embeddings"

        # Position embedding (serial).
        self.position_embeddings = torch.nn.Embedding(max_sequence_length, self.hidden_size)
        self._position_embeddings_key = "position_embeddings"
        # Initialize the position embeddings.
        self.init_method(self.position_embeddings.weight)

        # Token type embedding.
        # Add this as an optional field that can be added through
        # method call so we can load a pretrain model without
        # token types and add them as needed.
        self._tokentype_embeddings_key = "tokentype_embeddings"
        if self.num_tokentypes > 0:
            self.tokentype_embeddings = torch.nn.Embedding(self.num_tokentypes, self.hidden_size)
            # Initialize the token-type embeddings.
            self.init_method(self.tokentype_embeddings.weight)
        else:
            self.tokentype_embeddings = None

        # Embeddings dropout
        self.embedding_dropout = torch.nn.Dropout(embedding_dropout_prob)
        if torch.distributed.get_rank() == 0:
            print("FINISH WORD EMBEDDING", self.word_embeddings)

    def zero_parameters(self):
        """Zero out all parameters in embedding."""
        self.word_embeddings.weight.data.fill_(0)
        self.word_embeddings.weight.shared = True
        self.position_embeddings.weight.data.fill_(0)
        self.position_embeddings.weight.shared = True
        if self.num_tokentypes > 0:
            self.tokentype_embeddings.weight.data.fill_(0)
            self.tokentype_embeddings.weight.shared = True

    def add_tokentype_embeddings(self, num_tokentypes):
        """Add token-type embedding. This function is provided so we can add
        token-type embeddings in case the pretrained model does not have it.
        This allows us to load the model normally and then add this embedding.
        """
        if self.tokentype_embeddings is not None:
            raise Exception("tokentype embeddings is already initialized")
        if torch.distributed.get_rank() == 0:
            print("adding embedding for {} tokentypes".format(num_tokentypes), flush=True)
        self.num_tokentypes = num_tokentypes
        self.tokentype_embeddings = torch.nn.Embedding(num_tokentypes, self.hidden_size)
        # Initialize the token-type embeddings.
        self.init_method(self.tokentype_embeddings.weight)

    def forward(self, input_ids, position_ids, tokentype_ids=None):
        # Embeddings.
        words_embeddings = self.word_embeddings(input_ids)
        position_embeddings = self.position_embeddings(position_ids)
        embeddings = words_embeddings + position_embeddings
        if tokentype_ids is not None:
            assert self.tokentype_embeddings is not None
            embeddings = embeddings + self.tokentype_embeddings(tokentype_ids)
        else:
            assert self.tokentype_embeddings is None

        # Dropout.
        embeddings = self.embedding_dropout(embeddings)

        return embeddings

    def state_dict_for_save_checkpoint(self, destination=None, prefix="", keep_vars=False):
        """For easy load."""

        state_dict_ = {}
        state_dict_[self._word_embeddings_key] = self.word_embeddings.state_dict(destination, prefix, keep_vars)
        state_dict_[self._position_embeddings_key] = self.position_embeddings.state_dict(destination, prefix, keep_vars)
        if self.num_tokentypes > 0:
            state_dict_[self._tokentype_embeddings_key] = self.tokentype_embeddings.state_dict(
                destination, prefix, keep_vars
            )

        return state_dict_

    def load_state_dict(self, state_dict, strict=True):
        """Customized load."""

        # Word embedding.
        if self._word_embeddings_key in state_dict:
            state_dict_ = state_dict[self._word_embeddings_key]
        else:
            # for backward compatibility.
            state_dict_ = {}
            for key in state_dict.keys():
                if "word_embeddings" in key:
                    state_dict_[key.split("word_embeddings.")[1]] = state_dict[key]
        self.word_embeddings.load_state_dict(state_dict_, strict=strict)

        # Position embedding.
        if self._position_embeddings_key in state_dict:
            state_dict_ = state_dict[self._position_embeddings_key]
        else:
            # for backward compatibility.
            state_dict_ = {}
            for key in state_dict.keys():
                if "position_embeddings" in key:
                    state_dict_[key.split("position_embeddings.")[1]] = state_dict[key]
        self.position_embeddings.load_state_dict(state_dict_, strict=strict)

        # Tokentype embedding.
        if self.num_tokentypes > 0:
            state_dict_ = {}
            if self._tokentype_embeddings_key in state_dict:
                state_dict_ = state_dict[self._tokentype_embeddings_key]
            else:
                # for backward compatibility.
                for key in state_dict.keys():
                    if "tokentype_embeddings" in key:
                        state_dict_[key.split("tokentype_embeddings.")[1]] = state_dict[key]
            if len(state_dict_.keys()) > 0:
                self.tokentype_embeddings.load_state_dict(state_dict_, strict=strict)
            else:
                print(
                    "***WARNING*** expected tokentype embeddings in the " "checkpoint but could not find it", flush=True
                )


class TransformerLanguageModel(MegatronModule):
    """Transformer language model.

    Arguments:
        transformer_hparams: transformer hyperparameters
        vocab_size: vocabulary size
        max_sequence_length: maximum size of sequence. This
                             is used for positional embedding
        embedding_dropout_prob: dropout probability for embeddings
        num_tokentypes: size of the token-type embeddings. 0 value
                        will ignore this embedding
    """

    def __init__(
        self,
        init_method,
        output_layer_init_method,
        encoder_attn_mask_type,
        num_tokentypes=0,
        add_encoder=True,
        add_decoder=False,
        decoder_attn_mask_type=AttnMaskType.causal,
        add_pooler=False,
        pre_process=True,
        post_process=True,
    ):
        super().__init__()
        args = get_args()

        self.pre_process = pre_process
        self.post_process = post_process
        self.hidden_size = args.hidden_size
        self.num_tokentypes = num_tokentypes
        self.init_method = init_method
        self.add_encoder = add_encoder
        self.encoder_attn_mask_type = encoder_attn_mask_type
        self.add_decoder = add_decoder
        self.decoder_attn_mask_type = decoder_attn_mask_type
        self.add_pooler = add_pooler
        self.encoder_hidden_state = None

        # Embeddings.
        if self.pre_process:
            self.embedding = Embedding(
                self.hidden_size,
                args.padded_vocab_size,
                args.max_position_embeddings,
                args.hidden_dropout,
                self.init_method,
                self.num_tokentypes,
            )
            self._embedding_key = "embedding"

        # Transformer.
        # Encoder (usually set to True, False if part of an encoder-decoder
        # architecture and in encoder-only stage).
        if self.add_encoder:
            self.encoder = ParallelTransformer(
                self.init_method,
                output_layer_init_method,
                self_attn_mask_type=self.encoder_attn_mask_type,
                pre_process=self.pre_process,
                post_process=self.post_process,
            )
            self._encoder_key = "encoder"
        else:
            self.encoder = None

        # Decoder (usually set to False, True if part of an encoder-decoder
        # architecture and in decoder-only stage).
        if self.add_decoder:
            # Temporary assertion until we verify correctness of pipeline parallelism
            # implementation of T5.
            assert (
                args.pipeline_model_parallel_size == 1
            ), "pipeline parallelism is not supported in the presence of decoder"
            self.decoder = ParallelTransformer(
                self.init_method,
                output_layer_init_method,
                layer_type=LayerType.decoder,
                self_attn_mask_type=self.decoder_attn_mask_type,
                pre_process=self.pre_process,
                post_process=self.post_process,
            )
            self._decoder_key = "decoder"
        else:
            self.decoder = None

        if self.post_process:
            # Pooler.
            if self.add_pooler:
                self.pooler = Pooler(self.hidden_size, self.init_method)
                self._pooler_key = "pooler"

    def set_input_tensor(self, input_tensor):
        """ See megatron.model.transformer.set_input_tensor()"""

        # This is usually handled in schedules.py but some inference code still
        # gives us non-lists or None
        if not isinstance(input_tensor, list):
            input_tensor = [input_tensor]

        if self.add_encoder and self.add_decoder:
            assert (
                len(input_tensor) == 1
            ), "input_tensor should only be length 1 for stage with both encoder and decoder"
            self.encoder.set_input_tensor(input_tensor[0])
        elif self.add_encoder:
            assert len(input_tensor) == 1, "input_tensor should only be length 1 for stage with only encoder"
            self.encoder.set_input_tensor(input_tensor[0])
        elif self.add_decoder:
            if len(input_tensor) == 2:
                self.decoder.set_input_tensor(input_tensor[0])
                self.encoder_hidden_state = input_tensor[1]
            elif len(input_tensor) == 1:
                self.decoder.set_input_tensor(None)
                self.encoder_hidden_state = input_tensor[0]
            else:
                raise Exception("input_tensor must have either length 1 or 2")
        else:
            raise Exception("Stage must have at least either encoder or decoder")

    def forward(
        self,
        enc_input_ids,
        enc_position_ids,
        enc_attn_mask,
        dec_input_ids=None,
        dec_position_ids=None,
        dec_attn_mask=None,
        enc_dec_attn_mask=None,
        tokentype_ids=None,
        set_inference_key_value_memory=False,
        inference_max_sequence_len=None,
        pooling_sequence_index=0,
        enc_hidden_states=None,
        output_enc_hidden=False,
    ):

        # Encoder embedding.
        if self.pre_process:
            encoder_input = self.embedding(enc_input_ids, enc_position_ids, tokentype_ids=tokentype_ids)
        else:
            encoder_input = None

        # Run encoder.
        if enc_hidden_states is None:
            if self.encoder is not None:
                encoder_output = self.encoder(
                    encoder_input,
                    enc_attn_mask,
                    set_inference_key_value_memory=set_inference_key_value_memory,
                    inference_max_sequence_len=inference_max_sequence_len,
                )
            else:
                encoder_output = self.encoder_hidden_state
        else:
            encoder_output = enc_hidden_states.to(encoder_input.dtype)

        if self.post_process:
            if self.add_pooler:
                pooled_output = self.pooler(encoder_output, pooling_sequence_index)

        # output_enc_hidden refers to when we just need the encoder's
        # output. For example, it is helpful to compute
        # similarity between two sequences by average pooling
        if not self.add_decoder or output_enc_hidden:
            if self.add_pooler and self.post_process:
                return encoder_output, pooled_output
            else:
                return encoder_output

        # Decoder embedding.
        if self.pre_process:
            decoder_input = self.embedding(dec_input_ids, dec_position_ids)
        else:
            decoder_input = None

        # Run decoder.
        decoder_output = self.decoder(
            decoder_input,
            dec_attn_mask,
            encoder_output=encoder_output,
            enc_dec_attn_mask=enc_dec_attn_mask,
            set_inference_key_value_memory=set_inference_key_value_memory,
            inference_max_sequence_len=inference_max_sequence_len,
        )

        if self.add_pooler and self.post_process:
            return decoder_output, encoder_output, pooled_output
        else:
            return decoder_output, encoder_output

    def state_dict_for_save_checkpoint(self, destination=None, prefix="", keep_vars=False):
        """For easy load."""

        state_dict_ = {}
        if self.pre_process:
            state_dict_[self._embedding_key] = self.embedding.state_dict_for_save_checkpoint(
                destination, prefix, keep_vars
            )
        if self.add_encoder:
            state_dict_[self._encoder_key] = self.encoder.state_dict_for_save_checkpoint(destination, prefix, keep_vars)
        if self.post_process:
            if self.add_pooler:
                state_dict_[self._pooler_key] = self.pooler.state_dict_for_save_checkpoint(
                    destination, prefix, keep_vars
                )
        if self.add_decoder:
            state_dict_[self._decoder_key] = self.decoder.state_dict_for_save_checkpoint(destination, prefix, keep_vars)

        return state_dict_

    def load_state_dict(self, state_dict, strict=True):
        """Customized load."""

        # Embedding.
        if self.pre_process:
            if self._embedding_key in state_dict:
                state_dict_ = state_dict[self._embedding_key]
            else:
                # for backward compatibility.
                state_dict_ = {}
                for key in state_dict.keys():
                    if "_embeddings" in key:
                        state_dict_[key] = state_dict[key]
            self.embedding.load_state_dict(state_dict_, strict=strict)

        # Encoder.
        if self.add_encoder:
            if self._encoder_key in state_dict:
                state_dict_ = state_dict[self._encoder_key]
            # For backward compatibility.
            elif "transformer" in state_dict:
                state_dict_ = state_dict["transformer"]
            else:
                # For backward compatibility.
                state_dict_ = {}
                for key in state_dict.keys():
                    if "transformer." in key:
                        state_dict_[key.split("transformer.")[1]] = state_dict[key]

            # For backward compatibility.
            state_dict_self_attention = {}
            for key in state_dict_.keys():
                if ".attention." in key:
                    state_dict_self_attention[key.replace(".attention.", ".self_attention.")] = state_dict_[key]
                else:
                    state_dict_self_attention[key] = state_dict_[key]
            state_dict_ = state_dict_self_attention

            self.encoder.load_state_dict(state_dict_, strict=strict)

        # Pooler.
        if self.post_process:
            if self.add_pooler:
                assert "pooler" in state_dict, "could not find data for pooler in the checkpoint"
                self.pooler.load_state_dict(state_dict[self._pooler_key], strict=strict)
        # Decoder.
        if self.add_decoder:
            assert "decoder" in state_dict, "could not find data for pooler in the checkpoint"
            self.decoder.load_state_dict(state_dict[self._decoder_key], strict=strict)


def post_language_model_processing(lm_output, labels, logit_weights, parallel_output, fp16_lm_cross_entropy):
    # Output.
    output = parallel_lm_logits(lm_output, logit_weights, parallel_output)

    if labels is None:
        return output
    else:
        if fp16_lm_cross_entropy:
            assert output.dtype == torch.half
            loss = tensor_parallel.vocab_parallel_cross_entropy(output, labels)
        else:
            loss = tensor_parallel.vocab_parallel_cross_entropy(output.float(), labels)
        return loss

def module_size(m: torch.nn.Module, only_trainable: bool = False):
    """
    returns the total number of parameters used by `m` (only counting
    shared parameters once); if `only_trainable` is True, then only
    includes parameters with `requires_grad = True`
    """
    parameters = list(m.parameters())
    if only_trainable:
        parameters = [p for p in parameters if p.requires_grad]
    unique = {p.data_ptr(): p for p in parameters}.values()
    return sum(p.numel() for p in unique)

class GPTModel(MegatronModule):
    """GPT-2 Language model."""

    def __init__(self, num_tokentypes=0, parallel_output=True, pre_process=True, post_process=True, cpu_offload=False):
        super(GPTModel, self).__init__()
        args = get_args()
        self.parallel_output = parallel_output
        self.pre_process = pre_process
        self.post_process = post_process
        self.fp16_lm_cross_entropy = args.fp16_lm_cross_entropy
        self.cpu_offload = cpu_offload
        self.language_model, self._language_model_key = get_language_model(
            num_tokentypes=num_tokentypes,
            add_pooler=False,
            encoder_attn_mask_type=AttnMaskType.causal,
            init_method=init_method_normal(args.init_method_std),
            scaled_init_method=scaled_init_method_normal(args.init_method_std, args.num_layers),
            pre_process=self.pre_process,
            post_process=self.post_process,
        )

        self.initialize_word_embeddings(init_method_normal)

    def set_input_tensor(self, input_tensor):
        """See megatron.model.transformer.set_input_tensor()"""
        self.language_model.set_input_tensor(input_tensor)

    def forward(
        self,
        input_ids,
        position_ids,
        attention_mask,
        labels=None,
        tokentype_ids=None,
        set_inference_key_value_memory=False,
        inference_max_sequence_len=None,
    ):

        with torch.autograd.graph.save_on_cpu() if self.cpu_offload else contextlib.nullcontext():
            lm_output = self.language_model(
                input_ids,
                position_ids,
                attention_mask,
                set_inference_key_value_memory=set_inference_key_value_memory,
                inference_max_sequence_len=inference_max_sequence_len,
            )

            if self.post_process:
                return post_language_model_processing(
                    lm_output, labels, self.word_embeddings_weight(), self.parallel_output, self.fp16_lm_cross_entropy
                )
            else:
                return lm_output

    def state_dict_for_save_checkpoint(self, destination=None, prefix="", keep_vars=False):

        state_dict_ = {}
        state_dict_[self._language_model_key] = self.language_model.state_dict_for_save_checkpoint(
            destination, prefix, keep_vars
        )
        # Save word_embeddings.
        if self.post_process and not self.pre_process:
            state_dict_[self._word_embeddings_for_head_key] = self.word_embeddings.state_dict(
                destination, prefix, keep_vars
            )
        return state_dict_

    def load_state_dict(self, state_dict, strict=True):
        """Customized load."""

        # Load word_embeddings.
        if self.post_process and not self.pre_process:
            self.word_embeddings.load_state_dict(state_dict[self._word_embeddings_for_head_key], strict=strict)
        if self._language_model_key in state_dict:
            state_dict = state_dict[self._language_model_key]
        self.language_model.load_state_dict(state_dict, strict=strict)

def gpt_model_provider(pre_process=True, post_process=False, cpu_offload=False):
    model = GPTModel(num_tokentypes=0, parallel_output=True, pre_process=pre_process, post_process=post_process, cpu_offload=cpu_offload)
    if torch.distributed.get_rank() == 0:
        init_dict = {"pre_process":pre_process, "post_process":post_process, "cpu_offload":cpu_offload}
        print("Initialized GPT-2 w/:", json.dumps(init_dict))
        n_params = module_size(model) * parallel_state.get_tensor_model_parallel_world_size() * parallel_state.get_pipeline_model_parallel_world_size()
        print("Number of Parameters:", n_params)
    return model



# coding=utf-8
# Copyright (c) 2021, NVIDIA CORPORATION.  All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for pipeline model parallel."""
from typing import Optional, List, Union

import torch
from torch.nn.parallel import DistributedDataParallel

from apex.multi_tensor_apply import multi_tensor_applier
from apex.transformer import parallel_state
from apex.transformer.enums import ModelType
from apex.transformer.microbatches import build_num_microbatches_calculator
from apex.transformer.pipeline_parallel._timers import _Timers
if multi_tensor_applier.available:
    import amp_C


_GLOBAL_ARGS = None
_GLOBAL_NUM_MICROBATCHES_CALCULATOR = None
_GLOBAL_TOKENIZER = None
_GLOBAL_TENSORBOARD_WRITER = None
_GLOBAL_AUTORESUME = None
_GLOBAL_TIMERS = None


Shape = Union[List[int], torch.Size]


def listify_model(model: Union[torch.nn.Module, List[torch.nn.Module]]) -> List[torch.nn.Module]:
    if isinstance(model, list):
        return model
    return [model]


def _ensure_var_is_initialized(var, name):
    """Make sure the input variable is not None."""
    assert var is not None, "{} is not initialized.".format(name)


def _ensure_var_is_not_initialized(var, name):
    """Make sure the input variable is not None."""
    assert var is None, "{} is already initialized.".format(name)


def setup_microbatch_calculator(
        rank: int,
        rampup_batch_size: Optional[List[int]],
        global_batch_size: int,
        micro_batch_size: int,
        data_parallel_size: int,
) -> None:
    global _GLOBAL_NUM_MICROBATCHES_CALCULATOR
    _ensure_var_is_not_initialized(_GLOBAL_NUM_MICROBATCHES_CALCULATOR, 'num microbatches calculator')

    _GLOBAL_NUM_MICROBATCHES_CALCULATOR = build_num_microbatches_calculator(
        rank, rampup_batch_size, global_batch_size, micro_batch_size, data_parallel_size)


def _reconfigure_microbatch_calculator(
        rank: int,
        rampup_batch_size: Optional[List[int]],
        global_batch_size: int,
        micro_batch_size: int,
        data_parallel_size: int,
) -> None:
    if torch.distributed.get_rank() == 0:
        import warnings
        warnings.warn("This function is only for unittest")
    global _GLOBAL_NUM_MICROBATCHES_CALCULATOR

    _GLOBAL_NUM_MICROBATCHES_CALCULATOR = build_num_microbatches_calculator(
        rank, rampup_batch_size, global_batch_size, micro_batch_size, data_parallel_size)


def get_micro_batch_size():
    return _GLOBAL_NUM_MICROBATCHES_CALCULATOR.micro_batch_size


def get_num_microbatches():
    return _GLOBAL_NUM_MICROBATCHES_CALCULATOR.get()


def get_current_global_batch_size():
    return _GLOBAL_NUM_MICROBATCHES_CALCULATOR.get_current_global_batch_size()


def update_num_microbatches(consumed_samples, consistency_check=True):
    _GLOBAL_NUM_MICROBATCHES_CALCULATOR.update(consumed_samples, consistency_check)


# note (mkozuki): Comment out in favor of `get_kth_microbatch`
def _split_batch_into_microbatch(
        batch: List[torch.Tensor],
        *,
        _micro_batch_size: Optional[int] = None,
        _global_batch_size: Optional[int] = None,
) -> List[List[torch.Tensor]]:
    micro_batch_size = _micro_batch_size
    global_batch_size = _global_batch_size
    if micro_batch_size is None:
        micro_batch_size = get_micro_batch_size()
    if global_batch_size is None:
        global_batch_size = get_current_global_batch_size()
    for i in range(0, global_batch_size, micro_batch_size):
        yield [x[i * micro_batch_size:(i + 1) * micro_batch_size] for x in batch]


# TODO(mkozuki): Support non-tensor local minibatches?
def get_kth_microbatch(batch: Optional[List[torch.Tensor]], k: int) -> List[torch.Tensor]:
    """Create a list of microbatches from a list of local minibatches.

    This function creates a list of `k`th microbatches from a list of local minibatches.
    `a local minibatch` consists of `global_batch_size / data_parallel_size` samples.
    """
    if batch is None:
        return batch
    micro_batch_size = get_micro_batch_size()
    start = k * micro_batch_size
    end = start + micro_batch_size
    microbatch = list()
    for x in batch:
        size = x.size(0)
        assert size > start and size >= end
        microbatch.append(x[start:end])
    assert len(microbatch) > 0
    return microbatch


def get_autoresume():
    return _GLOBAL_AUTORESUME


def _set_timers():
    """Initialize timers."""
    global _GLOBAL_TIMERS
    _ensure_var_is_not_initialized(_GLOBAL_TIMERS, "timers")
    _GLOBAL_TIMERS = _Timers()


def get_timers():
    """Return timers."""
    _ensure_var_is_initialized(_GLOBAL_TIMERS, "timers")
    return _GLOBAL_TIMERS


def print_rank_0(message: str) -> None:
    """If distributed is initialized, print only on rank 0."""
    if torch.distributed.is_initialized():
        if torch.distributed.get_rank() == 0:
            print(message, flush=True)
    else:
        print(message, flush=True)


def is_last_rank():
    return torch.distributed.get_rank() == (torch.distributed.get_world_size() - 1)


def print_rank_last(message):
    """If distributed is initialized, print only on last rank."""
    if torch.distributed.is_initialized():
        if is_last_rank():
            print(message, flush=True)
    else:
        print(message, flush=True)


def param_is_not_shared(param: torch.nn.Parameter) -> bool:
    return getattr(param, "shared", False)


def unwrap_model(model, module_instances=(DistributedDataParallel,)):
    return_list = True
    if not isinstance(model, list):
        model = [model]
        return_list = False
    unwrapped_model = []
    for model_module in model:
        while isinstance(model_module, module_instances):
            model_module = model_module.module
        unwrapped_model.append(model_module)
    if not return_list:
        return unwrapped_model[0]
    return unwrapped_model


def get_model_type(
        model: torch.nn.Module,
) -> ModelType:
    """Get `model_type` of `model`.

    If ``model`` doesn't have ``model_type`` attribute, return ``ModelType.encoder_or_decoder``.

    Args:
        model
    """
    return getattr(unwrap_model(model), "model_type", ModelType.encoder_or_decoder)


def calc_params_l2_norm(model: torch.nn.Module, bf16: bool):
    """Calculate l2 norm of parameters """
    # args = get_args()
    if not isinstance(model, list):
        model = [model]
    # Remove duplicate params.
    params_data = []
    for model_ in model:
        for param in model_.parameters():
            is_not_shared = param_is_not_shared(param)
            is_not_tp_duplicate = parallel_state.param_is_not_tensor_parallel_duplicate(param)
            if is_not_shared and is_not_tp_duplicate:
                if bf16:
                    params_data.append(param.data.float())
                else:
                    params_data.append(param.data)
    # Calculate norm
    dummy_overflow_buf = torch.cuda.IntTensor([0])
    norm, _ = multi_tensor_applier(
        amp_C.multi_tensor_l2norm, dummy_overflow_buf, [params_data], False  # no per-parameter norm
    )
    norm_2 = norm * norm
    # Sum across all model-parallel GPUs.
    torch.distributed.all_reduce(
        norm_2, op=torch.distributed.ReduceOp.SUM, group=parallel_state.get_model_parallel_group()
    )
    return norm_2.item() ** 0.5


def average_losses_across_data_parallel_group(losses):
    """Reduce a tensor of losses across all GPUs."""
    averaged_losses = torch.cat([loss.clone().detach().view(1) for loss in losses])
    torch.distributed.all_reduce(averaged_losses, group=parallel_state.get_data_parallel_group())
    averaged_losses = averaged_losses / torch.distributed.get_world_size(
        group=parallel_state.get_data_parallel_group()
    )

    return averaged_losses


def report_memory(name):
    """Simple GPU memory report."""
    mega_bytes = 1024.0 * 1024.0
    string = name + " memory (MB)"
    string += " | allocated: {}".format(torch.cuda.memory_allocated() / mega_bytes)
    string += " | max allocated: {}".format(torch.cuda.max_memory_allocated() / mega_bytes)
    string += " | reserved: {}".format(torch.cuda.memory_reserved() / mega_bytes)
    string += " | max reserved: {}".format(torch.cuda.max_memory_reserved() / mega_bytes)
    if parallel_state.get_data_parallel_rank() == 0:
        print("[Rank {}] {}".format(torch.distributed.get_rank(), string), flush=True)


def print_params_min_max_norm(optimizer, iteration):
    """Print min, max, and norm of all parameters."""
    index = 0
    rank = torch.distributed.get_rank()
    string = "iteration, rank, index, tensor-model-parallel, min, max, norm\n"
    optimizer_ = optimizer.optimizer
    for param_group in optimizer_.param_groups:
        for param in param_group["params"]:
            index += 1
            min_ = param.data.min()
            max_ = param.data.max()
            norm = torch.linalg.norm(param.data)
            string += "{:7d}, {:4d}, {:4d}, {:2d}, ".format(
                iteration, rank, index, int(param.tensor_model_parallel)
            )
            string += "{:.6E}, {:.6E}, {:.6E}\n".format(min_, max_, norm)
    print(string, flush=True)


# NOTE (mkozuki): APEX doesn't have anything equivalent for
# `_GLOBAL_ADLR_AUTORESUME` like Megatron-LM.
# def check_adlr_autoresume_termination(iteration, model, optimizer, lr_scheduler, save: bool):
#     """Check for autoresume signal and exit if it is received."""
#     from apex.ppu.checkpointing import save_checkpoint
#
#     autoresume = get_adlr_autoresume()
#     # Add barrier to ensure consistency.
#     torch.distributed.barrier()
#     if autoresume.termination_requested():
#         if save:
#             save_checkpoint(iteration, model, optimizer, lr_scheduler)
#         print_rank_0(">>> autoresume termination request found!")
#         if torch.distributed.get_rank() == 0:
#             autoresume.request_resume()
#         print_rank_0(">>> training terminated. Returning")
#         sys.exit(0)


def get_ltor_masks_and_position_ids(
    data, eod_token, reset_position_ids, reset_attention_mask, eod_mask_loss
):
    """Build masks and position id for left to right model."""

    # Extract batch size and sequence length.
    micro_batch_size, seq_length = data.size()

    # Attention mask (lower triangular).
    if reset_attention_mask:
        att_mask_batch = micro_batch_size
    else:
        att_mask_batch = 1
    attention_mask = torch.tril(
        torch.ones((att_mask_batch, seq_length, seq_length), device=data.device)
    ).view(att_mask_batch, 1, seq_length, seq_length)

    # Loss mask.
    loss_mask = torch.ones(data.size(), dtype=torch.float, device=data.device)
    if eod_mask_loss:
        loss_mask[data == eod_token] = 0.0

    # Position ids.
    position_ids = torch.arange(seq_length, dtype=torch.long, device=data.device)
    position_ids = position_ids.unsqueeze(0).expand_as(data)
    # We need to clone as the ids will be modifed based on batch index.
    if reset_position_ids:
        position_ids = position_ids.clone()

    if reset_position_ids or reset_attention_mask:
        # Loop through the batches:
        for b in range(micro_batch_size):

            # Find indecies where EOD token is.
            eod_index = position_ids[b, data[b] == eod_token]
            # Detach indecies from positions if going to modify positions.
            if reset_position_ids:
                eod_index = eod_index.clone()

            # Loop through EOD indecies:
            prev_index = 0
            for j in range(eod_index.size()[0]):
                i = eod_index[j]
                # Mask attention loss.
                if reset_attention_mask:
                    attention_mask[b, 0, (i + 1) :, : (i + 1)] = 0
                # Reset positions.
                if reset_position_ids:
                    position_ids[b, (i + 1) :] -= i + 1 - prev_index
                    prev_index = i + 1

    # Convert attention mask to binary:
    attention_mask = attention_mask < 0.5

    return attention_mask, loss_mask, position_ids



import os
import torch
import json
import string
import time
try:
    from .permutation_search_kernels import accelerated_search_for_good_permutation, sum_after_2_to_4
    print("[ASP][Info] permutation_search_kernels can be imported.")
except ImportError:
    print("[ASP][Warning] permutation_search_kernels cannot be imported.")
    print("[ASP][Warning] If you want to accelerate the permutation search process by GPU, please build APEX by following the instructions at https://github.com/NVIDIA/apex/blob/master/apex/contrib/sparsity/README.md")

def convert_fx_node_name(fx_node_name):
    converted_fx_node_name = fx_node_name
    converted_fx_node_name = converted_fx_node_name.replace('_', '.')
    return converted_fx_node_name

def get_node_parent_children(fx_node):
    # get node parent list, and convert node name to module name
    node_parent_name_converted = []
    if len(fx_node.all_input_nodes) > 0:
        node_parent = fx_node.all_input_nodes
        for item in node_parent:
            converted_item = convert_fx_node_name(item.name)
            node_parent_name_converted.append(converted_item)
    else:
        node_parent = list('None')
        node_parent_name_converted.append('None')
    # get node children list, and convert node name to module name
    node_children_name_converted = []
    if len(list(fx_node.users.keys())) > 0:
        node_children = list(fx_node.users.keys())
        for item in node_children:
            converted_item = convert_fx_node_name(item.name)
            node_children_name_converted.append(converted_item)
    else:
        node_children = list('None')
        node_children_name_converted.append('None')
    return node_parent_name_converted, node_children_name_converted


class Permutation:
    __model = None
    __sparse_parameters = []
    __allow_permutation = False
    __all_parameters = []
    __save_permutation_graph = False
    __permutation_output_dir = ''

    @classmethod
    def set_permutation_params_from_asp(cls, model, sparse_parameters, all_parameters):
        """This function is used to set the permutation needed parameters from ASP class."""
        print("\n[set_permutation_params_from_asp] Set permutation needed parameters")
        cls.__model = model
        cls.__sparse_parameters = sparse_parameters
        cls.__all_parameters = all_parameters

    @classmethod
    def set_identical_seed(cls, identical_seed=1):
        print("\n[set_identical_seed] Set the identical seed: {:} for all GPUs to make sure the same results generated in permutation search".format(identical_seed))
        torch.manual_seed(identical_seed)
        torch.cuda.manual_seed_all(identical_seed)
        import numpy as np
        import random
        np.random.seed(identical_seed)
        random.seed(identical_seed)
        torch.backends.cudnn.deterministic = True
        torch.backends.cudnn.benchmark = False

    @classmethod
    def set_permutation_saving_params(cls, allow_permutation=False, save_permutation_graph=False, permutation_output_dir='.'):
        """This function is used to set the permutation saving related parameters."""
        print("\n[permutation_lib][set_permutation_saving_param] Set permutation saving related parameters")
        cls.__allow_permutation = allow_permutation
        print("[set_permutation_saving_param]\t Allow permutation: {}".format(cls.__allow_permutation))
        cls.__save_permutation_graph = save_permutation_graph
        print("[set_permutation_saving_param]\t Save permutation graphs: {}".format(cls.__save_permutation_graph))
        cls.__permutation_output_dir = permutation_output_dir
        print("[set_permutation_saving_param]\t Permutation graphs saving dir: {}".format(cls.__permutation_output_dir))

    @classmethod
    def apply_offline_permutation(cls, model, fx_graph):
        """This function is used to offline permutation for each node according to the the whole network graph built with Torch.FX."""
        print("\n[apply_offline_permutation] Offline permutation for each node according to the the whole network graph built with Torch.FX")

        # Firstly, we should transfer the sparse mask to all-one dense mask
        cls.transfer_to_dense_mask()

        for node_name in fx_graph.keys():
            node_module_type = fx_graph.get(node_name).get('module_type')

            # check wheter the current layer can permute as plan, e.g., the flatten layer in VGG will change the shape and broke the permutation chain
            # only need to check the 'is_node_real_parents_K_permuted', because the 'is_node_real_parents_C_permuted' has no influence to the children
            node_real_parents = fx_graph.get(node_name).get('real_parents')
            is_node_real_parents_K_permuted = True
            if node_real_parents is not None:    # filter out the 'unique_siblings' item
                for real_parent_item in node_real_parents:
                    if fx_graph.get(real_parent_item).get('permutation_type') in ['K', 'KC']:
                        if fx_graph.get(real_parent_item).get('k_permuted') == 'False':
                            is_node_real_parents_K_permuted = False

            if fx_graph[node_name]['permutation_type'] == 'KC':    # intermediate Conv, FC
                C_permutation_sequence = cls.fetch_C_permutation_sequence_value(node_name, fx_graph)
                K_permutation_sequence = cls.fetch_K_permutation_sequence_value(node_name, fx_graph)
                print("\n[apply_offline_permutation] node_name: \'{:}\', node module type: \'{:}\', need to do offline permutation in K and C dims.".format(node_name, node_module_type))
                if is_node_real_parents_K_permuted == True:
                    fx_graph[node_name]['c_permuted'] = str(cls.apply_permutation_in_C_dim(node_name, C_permutation_sequence))
                    fx_graph[node_name]['k_permuted'] = str(cls.apply_permutation_in_K_dim(node_name, K_permutation_sequence))
                else:
                    print("[apply_offline_permutation][warning] node_name: \'{:}\', its real parents have trouble in permutation in K dim, so skip the offline permutation in C dim.".format(node_name, node_module_type))
                    fx_graph[node_name]['k_permuted'] = str(cls.apply_permutation_in_K_dim(node_name, K_permutation_sequence))
            elif fx_graph[node_name]['permutation_type'] == 'K':    # BN, first layer Conv/FC
                K_permutation_sequence = cls.fetch_K_permutation_sequence_value(node_name, fx_graph)
                print("\n[apply_offline_permutation] node_name: \'{:}\', node module type: \'{:}\', need to do offline permutation in K dim.".format(node_name, node_module_type))
                if is_node_real_parents_K_permuted == True:
                    fx_graph[node_name]['k_permuted'] = str(cls.apply_permutation_in_K_dim(node_name, K_permutation_sequence))
                else:    # for BN, if the previous Conv cannot do permutation in K dim, then no need to do permutation in K dim for this BN
                    print("[apply_offline_permutation][warning] node_name: \'{:}\', its real parents have trouble in permutation in K dim, so skip the offline permutation in K dim.".format(node_name, node_module_type))
            elif fx_graph[node_name]['permutation_type'] == 'C':    # last layer FC/Conv
                C_permutation_sequence = cls.fetch_C_permutation_sequence_value(node_name, fx_graph)
                print("\n[apply_offline_permutation] node_name: \'{:}\', node module type: \'{:}\', need to do offline permutation in C dim.".format(node_name, node_module_type))
                if is_node_real_parents_K_permuted == True:
                    fx_graph[node_name]['c_permuted'] = str(cls.apply_permutation_in_C_dim(node_name, C_permutation_sequence))
                else:
                    print("[apply_offline_permutation][warning] node_name: \'{:}\', its real parents have trouble in permutation in K dim, so skip the offline permutation in C dim.".format(node_name, node_module_type))

        if cls.__save_permutation_graph:
            cls.save_graph_to_json(fx_graph, save_dumped_graph_path_with_name=os.path.join(cls.__permutation_output_dir, './model_graph_apply_offline_permutation.json'))    # save the intermediate graph as JSON file for debugging
        return fx_graph

    @classmethod
    def transfer_to_dense_mask(cls):
        """Call this method to transfer the sparse mask to all-one dense mask."""
        with torch.no_grad():
            for module_name, module, p_name, p, mask, pruned in cls.__sparse_parameters:
                mask.fill_(1)

    @classmethod
    def fetch_C_permutation_sequence_value(cls, node_name, fx_graph):
        """This function is used to fetch the permutation sequence value in C dim from the unique_siblings record."""
        # C_permutation_sequence is the corresponding 'permutation_sequence' value stored in the fx_graph.get('unique_siblings') item which contains node_name
        unique_siblings_groups = fx_graph.get('unique_siblings').get('name')
        unique_siblings_groups_permutation_sequence = fx_graph.get('unique_siblings').get('permutation_sequence')
        item_index = 0
        fetched_C_permutation_sequence = []
        for item in unique_siblings_groups:
            if node_name in item:
                fetched_C_permutation_sequence = unique_siblings_groups_permutation_sequence[item_index]
            item_index = item_index + 1
        return fetched_C_permutation_sequence

    @classmethod
    def fetch_K_permutation_sequence_value(cls, node_name, fx_graph):
        """This function is used to fetch the permutation sequence value in K dim from the unique_siblings record."""
        # K_permutation_sequence is its real_children's corresponding 'permutation_sequence' value stored in the fx_graph.get('unique_siblings') item which contains real_children name
        # we have the assumption that all the real children are in one unique_sibling group, so should share the same permutation_sequence value
        unique_siblings_groups = fx_graph.get('unique_siblings').get('name')
        unique_siblings_groups_permutation_sequence = fx_graph.get('unique_siblings').get('permutation_sequence')
        node_real_children = fx_graph.get(node_name).get('real_children')
        fetched_K_permutation_sequence = []
        if len(node_real_children) > 0:
            node_representative_child = node_real_children[0]
            fetched_K_permutation_sequence = cls.fetch_C_permutation_sequence_value(node_representative_child, fx_graph)
        return fetched_K_permutation_sequence

    @classmethod
    def apply_permutation_in_C_dim(cls, node_name, permutation_sequence):
        """This function is used to permutation for a node in C dim. (Only need to handle the weight of the node) """
        print("[apply_permutation_in_C_dim] Permutation for node: \'{:}\' in C dim".format(node_name))
        if len(permutation_sequence) == 0:
            print("[apply_permutation_in_C_dim] the permutation sequence is empty, fail to apply permutation in C dim.")
            return False
        is_node_in_sparse_parameters = False
        success_permutation = False
        for module_name, module, p_name, p, mask, pruned in cls.__sparse_parameters:
            processed_module_name = ''.join(c for c in module_name if c not in string.punctuation).lower()
            processed_node_name = ''.join(c for c in node_name if c not in string.punctuation).lower()
            distributed_node_name = 'module.' + node_name
            processed_distributed_node_name = 'module.' + processed_node_name
            if (module_name == node_name) or (module_name == distributed_node_name) or (processed_module_name == processed_node_name) or (processed_module_name == processed_distributed_node_name):    # Inception-V3, module_name: Conv2d_2a_3x3.conv, node_name: conv2d.1a.3x3.conv
                print("[apply_permutation_in_C_dim] find the node: \'{:}\' in cls.__sparse_parameters, succeed to apply permutation in C dim.".format(node_name))
                is_node_in_sparse_parameters = True
                temp_weight = torch.zeros_like(p)
                temp_weight.copy_(p[:, permutation_sequence, ...])
                p.data.copy_(temp_weight)
                success_permutation = True
        if is_node_in_sparse_parameters == False:
            # A special case: if the node itself not in sparse_module_names but one of its real_siblings in sparse_module_names, then the node will not do the permutation search, but it may need to apply the offline permutation in C dim according to the searched permutation sequence from its real_siblings in sparse_module_names
            try:
                for module_name_from_all_parameters, module_from_all_parameters, p_name_from_all_parameters, p_from_all_parameters in cls.__all_parameters:
                    if ((node_name == module_name_from_all_parameters) or ('module.' + node_name == module_name_from_all_parameters)) and p_name_from_all_parameters == "weight":
                        print("[apply_permutation_in_C_dim] cannot find the node: \'{:}\' in cls.__sparse_parameters, but can find in cls.__all_parameters.".format(node_name))
                        temp_weight = torch.zeros_like(p_from_all_parameters)
                        temp_weight.copy_(p_from_all_parameters[:, permutation_sequence, ...])
                        p_from_all_parameters.data.copy_(temp_weight)
                        success_permutation = True
                        print("[apply_permutation_in_C_dim] cannot find the node: \'{:}\' in cls.__sparse_parameters, after trying with cls.__all_parameters, succeed to apply permutation in C dim.".format(node_name))
            except:
                success_permutation = False
                print("[apply_permutation_in_C_dim] cannot find the node: \'{:}\' in cls.__sparse_parameters, after trying with cls.__all_parameters, still fail to apply permutation in C dim.".format(node_name))
        return success_permutation

    @classmethod
    def apply_permutation_in_K_dim(cls, node_name, permutation_sequence):
        """This function is used to permutation for a node in K dim. (Need to handle the weight/bias/running_mean/running_var of the node)"""
        print("[apply_permutation_in_K_dim] Permutation for node: \'{:}\' in K dim".format(node_name))
        if len(permutation_sequence) == 0:
            print("[apply_permutation_in_K_dim] the permutation sequence is empty, fail to apply permutation in K dim.")
            return False
        is_node_in_all_parameters = False
        success_permutation = False
        for module_name, module, p_name, p in cls.__all_parameters:
            processed_module_name = ''.join(c for c in module_name if c not in string.punctuation).lower()
            processed_node_name = ''.join(c for c in node_name if c not in string.punctuation).lower()
            distributed_node_name = 'module.' + node_name
            processed_distributed_node_name = 'module.' + processed_node_name
            if (module_name == node_name) or (module_name == distributed_node_name) or (processed_module_name == processed_node_name) or (processed_module_name == processed_distributed_node_name):    # Inception-V3, module_name: Conv2d_2a_3x3.conv, node_name: conv2d.1a.3x3.conv
                print("[apply_permutation_in_K_dim] find the node: \'{:}\' with \'{:}\' in cls.__all_parameters, may succeed to apply permutation in K dim.".format(node_name, p_name))
                is_node_in_all_parameters = True
                temp_weight = torch.zeros_like(p)
                if p.shape[0] != len(permutation_sequence):
                    print("[apply_permutation_in_K_dim][warning] the node: \'{:}\' with shape: \'{:}\', cannot match the size of permutation sequence with len: \'{:}\', fail to apply permutation in K dim.".format(node_name, p.shape, len(permutation_sequence)))
                    success_permutation = False
                else:
                    print("[apply_permutation_in_K_dim] the node: \'{:}\' with shape: \'{:}\', can match the size of permutation sequence with len: \'{:}\', succeed to apply permutation in K dim.".format(node_name, p.shape, len(permutation_sequence)))
                    temp_weight.copy_(p[permutation_sequence, ...])
                    p.data.copy_(temp_weight)
                    success_permutation = True
        if is_node_in_all_parameters == False:
            print("[apply_permutation_in_K_dim] cannot find the node: \'{:}\' in cls.__all_parameters, fail to apply permutation in K dim.".format(node_name))
            success_permutation = False
        return success_permutation

    @classmethod
    def build_offline_permutation_graph(cls, model, dump_fx_graph=False, save_dumped_fx_graph='./model_offline_permutation_graph.json'):
        """This function is used to refine the whole network graph built with Torch.FX with some extra infomation needed for offline permutation."""
        print("\n[build_offline_permutation_graph] Further refine the model graph built by Torch.FX for offline permutation")
        # extract the output_dir, so all the intermediate fx_graph can be saved under that path
        extract_output_dir=os.path.split(save_dumped_fx_graph)[0]
        cls.__permutation_output_dir = extract_output_dir
        fx_graph, success_in_build_fx_graph = cls.build_fx_graph(model, dump_fx_graph=dump_fx_graph, save_dumped_fx_graph=save_dumped_fx_graph)
        if success_in_build_fx_graph:
            fx_graph_after_find_real_parents = cls.find_real_parents(fx_graph)
            fx_graph_after_find_real_children = cls.find_real_children(fx_graph_after_find_real_parents)
            fx_graph_after_find_real_siblings = cls.find_real_siblings(fx_graph_after_find_real_children)
            fx_graph_after_extract_all_unique_siblings = cls.extract_all_unique_siblings(fx_graph_after_find_real_siblings)
            fx_graph_after_init_permutation_flag = cls.init_permutation_flag(fx_graph_after_extract_all_unique_siblings)
            start_time_search_for_good_permutation = time.perf_counter()
            fx_graph_after_search_for_good_permutation = cls.search_for_good_permutation(fx_graph_after_init_permutation_flag)
            duration_search_for_good_permutation = time.perf_counter() - start_time_search_for_good_permutation
            print("\n[build_offline_permutation_graph] Take {:.4f} seconds to finish search_for_good_permutation function.".format(duration_search_for_good_permutation))
        else:
            fx_graph_after_search_for_good_permutation = {}
            return fx_graph_after_search_for_good_permutation, success_in_build_fx_graph

        # Please notice the apply_offline_permutation step cannot fold into the above search_for_good_permutation step.
        # Because the real_parent node needs to offline permutation in K direction according to the searched permutation sequence from its real_children.
        # However, when we search_for_good_permutation for the node, its real_children have not been handled by search_for_good_permutation.

        if cls.__save_permutation_graph:
            cls.save_graph_to_json(fx_graph_after_search_for_good_permutation, save_dumped_graph_path_with_name=os.path.join(cls.__permutation_output_dir, './model_graph_build_offline_permutation_graph.json'))    # save the intermediate graph as JSON file for debugging
        return fx_graph_after_search_for_good_permutation, success_in_build_fx_graph

    @classmethod
    def search_for_good_permutation(cls, fx_graph):
        """This function is used to:
        1. search for the good permutation sequence for each node weight, or each siblings_group weights by calling the permutation search kernels as ASP extension.
        2. add the searched permutation sequence for each node according to the whole network graph built with Torch.FX."""
        print("\n[search_for_good_permutation] Search for the good permutation sequence for each node according to the whole network graph built with Torch.FX")

        unique_siblings_groups = fx_graph.get('unique_siblings').get('name')
        unique_siblings_groups_module_type = fx_graph.get('unique_siblings').get('module_type')
        unique_siblings_groups_permutation_sequence = []
        item_index = 0
        for unique_siblings_group in unique_siblings_groups:    # loop through all unique siblings groups that must share a permutation sequence
            print("\n[search_for_good_permutation] this unique_siblings_group has {:} real siblings: \'{:}\', with module type: \'{:}\'.".format(len(unique_siblings_group), unique_siblings_group, unique_siblings_groups_module_type[item_index]))
            item_index = item_index + 1

            # concat the weight for layers in the same unique_siblings_group
            matrix_group = None
            for node_name in unique_siblings_group:
                node_module_type = fx_graph.get(node_name).get('module_type')
                print("[search_for_good_permutation] try to merge the weight for node: \'{:}\', with module type: \'{:}\'.".format(node_name, node_module_type))
                is_node_in_sparse_parameters = False
                node_weight = torch.zeros(0)
                for module_name, module, p_name, p, mask, pruned in cls.__sparse_parameters:
                    processed_module_name = ''.join(c for c in module_name if c not in string.punctuation).lower()
                    processed_node_name = ''.join(c for c in node_name if c not in string.punctuation).lower()
                    distributed_node_name = 'module.' + node_name
                    processed_distributed_node_name = 'module.' + processed_node_name
                    if (module_name == node_name) or (module_name == distributed_node_name) or (processed_module_name == processed_node_name) or (processed_module_name == processed_distributed_node_name):    # Inception-V3, module_name: Conv2d_2a_3x3.conv, node_name: conv2d.1a.3x3.conv
                        module_type_from_sparse_parameters = str(type(module))    # e.g. <class 'torch.nn.modules.conv.Conv2d'>
                        module_type_from_sparse_parameters = module_type_from_sparse_parameters[8:-2]
                        print("[search_for_good_permutation] find the node: \'{:}\' in cls.__sparse_parameters, module type match: \'{:}\'.".format(node_name, node_module_type==module_type_from_sparse_parameters))
                        is_node_in_sparse_parameters = True
                        node_weight = torch.zeros_like(p)
                        node_weight.copy_(p)
                        # Need to handle the concat for layers with different R & S
                        shape = node_weight.shape
                        # 1d-tensor
                        if len(shape) == 1:
                            node_weight = node_weight.view(1, shape[0])
                        # 2d-tensor (in, out)
                        elif len(shape) == 2:
                            node_weight = node_weight.view(shape[0], shape[1])
                        # 3d-tensor (batch, in, out)
                        elif len(shape) == 3:
                            node_weight = node_weight.view(shape[0]*shape[1], shape[2])
                        # 4d-tensor (in, out, h, w)
                        elif len(shape) == 4:
                            # convs
                            node_weight = node_weight.permute(2,3,0,1).contiguous().view(shape[2]*shape[3]*shape[0], shape[1])

                if is_node_in_sparse_parameters == False:
                    print("[search_for_good_permutation] cannot find the node: \'{:}\' in cls.__sparse_parameters, no need to merge its weight for permutation.".format(node_name))
                else:
                    if matrix_group == None:
                        matrix_group = node_weight
                    else:
                        try:
                            if matrix_group.dim() == node_weight.dim():
                                matrix_group = torch.cat((matrix_group, node_weight), dim=0)    # concat the weights in K dimension, and keep the same C dimension
                            else:    # e.g. when try to merge the Conv and FC layers
                                print("[search_for_good_permutation] matrix_group dim: {:} is not matched with node_weight dim: {:}.".format(matrix_group.dim(), node_weight.dim()))
                                print("[search_for_good_permutation] matrix_group shape: \'{:}\' is not matched with node_weight shape: \'{:}\'.".format(matrix_group.size(), node_weight.size()))
                                if matrix_group.dim() < node_weight.dim():
                                    while node_weight.dim() - matrix_group.dim() > 0:
                                        matrix_group = matrix_group.unsqueeze(matrix_group.dim())
                                else:
                                    while matrix_group.dim() - node_weight.dim() > 0:
                                        node_weight = node_weight.unsqueeze(node_weight.dim())
                                print("[search_for_good_permutation] matrix_group shape: \'{:}\' is now matched with node_weight shape: \'{:}\'.".format(matrix_group.size(), node_weight.size()))
                                matrix_group = torch.cat((matrix_group, node_weight), dim=0)    # concat the weights in K dimension, and keep the same C dimension
                        except:
                            print("[search_for_good_permutation][warning] cannot merge the weight for node: \'{:}\', with its weight shape: \'{:}\', the matrix_group shape: \'{:}\'.".format(node_name, node_weight.size(), matrix_group.size()))
                            continue
                    print("[search_for_good_permutation] have merged the weight for node: \'{:}\', with its weight shape: \'{:}\', the matrix_group shape: \'{:}\'.".format(node_name, node_weight.size(), matrix_group.size()))

            if matrix_group == None:    # cannot find the node: \'{:}\' in cls.__sparse_parameters
                input_channel_num = 0
                print("\n[search_for_good_permutation] init the all-zero list with length \'{:}\' for permutation search sequence of this unique_siblings_group.".format(input_channel_num))
                print("[search_for_good_permutation] no need to search the permutation_sequence for empty matrix_group.")
                permutation_sequence = [0 for n in range(input_channel_num)]
                unique_siblings_groups_permutation_sequence.append(permutation_sequence)
                continue
            else:
                input_channel_num = matrix_group.size()[1]
                print("\n[search_for_good_permutation] init the all-zero list with length \'{:}\' for permutation search sequence of this unique_siblings_group.".format(input_channel_num))
                permutation_sequence = [0 for n in range(input_channel_num)]

            # automatic check for skipping the permutation search process
            original_magnitude = (torch.abs(matrix_group)).sum(dtype=torch.float64)
            pruned_magnitude = sum_after_2_to_4(matrix_group.cpu().detach().numpy())
            diff_ratio = abs(original_magnitude - pruned_magnitude)/original_magnitude
            epsilon = 1e-3
            print("\n[search_for_good_permutation] Original element abs sum: {:}, Pruned element abs sum: {:}, Diff ratio: {:}".format(original_magnitude, pruned_magnitude, diff_ratio))
            if diff_ratio < epsilon:
                print("[search_for_good_permutation] Original element abs sum is almost same as the pruned element abs sum, further permutation search will not help, skipping!")
                print("[search_for_good_permutation] Change the all-zero permutation search sequence to a sequential permutation search sequence.")
                permutation_sequence = [n for n in range(input_channel_num)]
                unique_siblings_groups_permutation_sequence.append(permutation_sequence)
                continue
            else:
                print("[search_for_good_permutation] Original element abs sum is different from the pruned element abs sum, further permutation search will help, continue with the permutation search!")

            # call the permutation search CUDA kernels as ASP extension.
            # users can provide prefer search strategy by providing a valid 'search_options' as a dictionary,
            # or users can implement their customized 'accelerated_search_for_good_permutation' function.
            search_options = {}
            # No.1 Strategy: Exhaustive Search
            # search_options['strategy'] = 'exhaustive'
            # search_options['stripe_group_size'] = 8
            # search_options['escape_attempts'] = 100
            # No.2 Strategy: Progressive Channel Swap Search
            # search_options['strategy'] = 'progressive channel swap'
            # search_options['progressive_search_time_limit'] = 10
            # search_options['improvement_threshold'] = 1e-9
            # No.3 Strategy: User Defined Search
            # search_options['strategy'] = 'user defined'

            # permutation search time is too long for matrix_group with large channel num
            # change from Exhaustive Search to Progressive Channel Swap Search based on input matrix_group size
            if input_channel_num > 2048:
                search_options['strategy'] = 'progressive channel swap'
                search_options['progressive_search_time_limit'] = 120
                search_options['improvement_threshold'] = 1e-9
                print("[search_for_good_permutation] Change to Progressive Channel Swap Search with {} seconds limitation, because the {} is too large and will leading too long permutation search time with Exhaustive Search.".format(search_options['progressive_search_time_limit'], input_channel_num))

            start_time_accelerated_search_for_good_permutation = time.perf_counter()
            permutation_sequence = accelerated_search_for_good_permutation(matrix_group, options=search_options)
            duration_accelerated_search_for_good_permutation = time.perf_counter() - start_time_accelerated_search_for_good_permutation
            print("[search_for_good_permutation] Take {:.4f} seconds to finish accelerated_search_for_good_permutation function.".format(duration_accelerated_search_for_good_permutation))
            unique_siblings_groups_permutation_sequence.append(permutation_sequence)
        fx_graph['unique_siblings']['permutation_sequence'] = unique_siblings_groups_permutation_sequence

        if cls.__save_permutation_graph:
            cls.save_graph_to_json(fx_graph, save_dumped_graph_path_with_name=os.path.join(cls.__permutation_output_dir, './model_graph_search_for_good_permutation.json'))    # save the intermediate graph as JSON file for debugging
        return fx_graph

    @classmethod
    def init_permutation_flag(cls, fx_graph):
        """This function is used to init the permutation flag for each node according to the whole network graph built with Torch.FX."""
        print("\n[init_permutation_flag] Init the permutation flag for each node according to the whole network graph built with Torch.FX")
        sparse_module_names = []
        processed_sparse_module_names = []    # Inception-V3, module_name: Conv2d_2a_3x3.conv, node_name: conv2d.1a.3x3.conv
        for module_name, module, p_name, p, mask, pruned in cls.__sparse_parameters:
            sparse_module_names.append(module_name)
            processed_module_name = ''.join(c for c in module_name if c not in string.punctuation).lower()
            processed_sparse_module_names.append(processed_module_name)
        for node_name in fx_graph.keys():
            processed_node_name = ''.join(c for c in node_name if c not in string.punctuation).lower()
            distributed_node_name = 'module.' + node_name
            processed_distributed_node_name = 'module.' + processed_node_name
            node_module_type = fx_graph.get(node_name).get('module_type')
            if node_module_type in ['torch.nn.modules.conv.Conv2d', 'torch.nn.modules.linear.Linear']:
                node_parents = fx_graph.get(node_name).get('parents')
                node_children = fx_graph.get(node_name).get('children')
                node_real_parents = fx_graph.get(node_name).get('real_parents')
                node_real_children = fx_graph.get(node_name).get('real_children')
                node_groups_param = fx_graph.get(node_name).get('groups_param')
                is_node_real_children_in_sparse_parameters = False
                is_node_real_children_has_group_conv = False
                for real_child_item in node_real_children:
                    processed_real_child_item = ''.join(c for c in real_child_item if c not in string.punctuation).lower()
                    distributed_real_child_item = 'module.' + real_child_item
                    processed_distributed_real_child_item = 'module.' + processed_real_child_item
                    if (real_child_item in sparse_module_names) or (processed_real_child_item in processed_sparse_module_names) or (distributed_real_child_item in sparse_module_names) or (processed_distributed_real_child_item in processed_sparse_module_names):
                        is_node_real_children_in_sparse_parameters = True
                    if (fx_graph.get(real_child_item).get('groups_param') not in ['None', '1']):
                        is_node_real_children_has_group_conv = True
                is_node_real_parents_has_group_conv = False
                for real_parent_item in node_real_parents:
                    # notice: we assume the if one item of real_parents need to permute in C or K dim, then the corresponding flag should be set
                    # so for all items of real_parents, they may not share the same 'permutation_type' (e.g., one item is Group Conv, etc.)
                    # that's why we also need to judge the 'is_node_real_parents_has_group_conv'
                    if (fx_graph.get(real_parent_item).get('groups_param') not in ['None', '1']):
                        is_node_real_parents_has_group_conv = True
                # If the node itself is in sparse_module_names or one of its real_children in sparse_module_names, then it may need the offline permutation
                if ((node_name in sparse_module_names) or (processed_node_name in processed_sparse_module_names) or (distributed_node_name in sparse_module_names) or (processed_distributed_node_name in processed_sparse_module_names)) or (is_node_real_children_in_sparse_parameters == True):
                    if node_groups_param not in ['None', '1']:
                        # for Group Conv, disable the permutation in 'C' and 'K' dim
                        fx_graph[node_name]['permutation_type'] = 'None'
                    elif ('x' in node_parents) or ((node_name not in sparse_module_names) and (processed_node_name not in processed_sparse_module_names) and (distributed_node_name not in sparse_module_names) and (processed_distributed_node_name not in processed_sparse_module_names)):
                        # for the first (due to it is connected to 'x' node or itself is not in sparse_module_names) or not NVIDIA's TC compatiable Conv/FC, only permutate the K direction
                        if is_node_real_children_has_group_conv == False:
                            fx_graph[node_name]['permutation_type'] = 'K'
                            fx_graph[node_name]['k_permuted'] = 'False'
                        else:    # if node real_children contains Group Conv, disable the permutation for node in 'K' dim
                            fx_graph[node_name]['permutation_type'] = 'None'
                    elif ('output' in node_children) or (is_node_real_children_in_sparse_parameters == False):
                        # for the last (due to it is connected to 'output' node or to a node which is not in sparse_module_names) FC/Conv, only permutate the C direction
                        if is_node_real_parents_has_group_conv == False:
                            fx_graph[node_name]['permutation_type'] = 'C'
                            fx_graph[node_name]['c_permuted'] = 'False'
                        else:    # if node real_parents contains Group Conv, disable the permutation for node in 'C' dim
                            fx_graph[node_name]['permutation_type'] = 'None'
                    else:
                        if (is_node_real_parents_has_group_conv == False) and (is_node_real_children_has_group_conv == False):
                            fx_graph[node_name]['permutation_type'] = 'KC'
                            fx_graph[node_name]['k_permuted'] = 'False'
                            fx_graph[node_name]['c_permuted'] = 'False'
                        elif is_node_real_parents_has_group_conv == True:    # TODO: if node real_parents contains Group Conv, disable the permutation for node in 'C' dim
                            fx_graph[node_name]['permutation_type'] = 'K'
                            fx_graph[node_name]['k_permuted'] = 'False'
                        else:    # if node real_children contains Group Conv, disable the permutation for node in 'K' dim
                            fx_graph[node_name]['permutation_type'] = 'C'
                            fx_graph[node_name]['c_permuted'] = 'False'
                else:
                    fx_graph[node_name]['permutation_type'] = 'None'
            elif node_module_type in ['torch.nn.modules.batchnorm.BatchNorm2d']:
                node_real_parents = fx_graph.get(node_name).get('real_parents')
                is_node_real_parents_need_K_permutation = False
                is_node_real_parents_has_group_conv = False
                for real_parent_item in node_real_parents:
                    # notice: we assume the if one item of real_parents need to permute in K dim, then the corresponding flag should be set
                    # as in most of the cases, BN only follows one Conv, so it should be OK for now
                    if fx_graph.get(real_parent_item).get('permutation_type') in ['K', 'KC']:
                        is_node_real_parents_need_K_permutation = True
                    if (fx_graph.get(real_parent_item).get('groups_param') not in ['None', '1']):
                        is_node_real_parents_has_group_conv = True
                node_real_children = fx_graph.get(node_name).get('real_children')
                is_node_real_children_in_sparse_parameters = False
                for real_child_item in node_real_children:
                    processed_real_child_item = ''.join(c for c in real_child_item if c not in string.punctuation).lower()
                    distributed_real_child_item = 'module.' + real_child_item
                    processed_distributed_real_child_item = 'module.' + processed_real_child_item
                    if (real_child_item in sparse_module_names) or (processed_real_child_item in processed_sparse_module_names) or (distributed_real_child_item in sparse_module_names) or (processed_distributed_real_child_item in processed_sparse_module_names):
                        is_node_real_children_in_sparse_parameters = True
                # Firstly, we should make sure the BN is not in the last (due to it is connected to a FC/Conv node which is not in sparse_module_names), then:
                # If the real_parents of BN node are in sparse_module_names, then it may need the offline permutation
                # Or if the real_parents of BN node just needs to permute in K dim
                if (is_node_real_children_in_sparse_parameters == True) and (is_node_real_parents_need_K_permutation == True):
                    if (is_node_real_parents_has_group_conv == False) and (is_node_real_parents_need_K_permutation == True):
                        fx_graph[node_name]['permutation_type'] = 'K'
                        fx_graph[node_name]['k_permuted'] = 'False'
                    else:    # if node real_parents contains Group Conv or does not need permutation in 'K' dim, disable the permutation for node in 'K' dim
                        fx_graph[node_name]['permutation_type'] = 'None'
                else:
                    fx_graph[node_name]['permutation_type'] = 'None'
            else:
                fx_graph[node_name]['permutation_type'] = 'None'

        # A special case: if the node itself not in sparse_module_names but one of its real_siblings in sparse_module_names, then the node will not do the permutation search, but it may need to apply the offline permutation in C dim according to the searched permutation sequence from its real_siblings in sparse_module_names
        # We make it as the post-processing, because if we add this to the previous logic, will make it too complex
        # Post-processing Step No.1:
        print("\n[init_permutation_flag] Post-processing Step No.1.")
        node_change_permutation_due_to_siblings = []
        for node_name in fx_graph.keys():
            node_real_siblings = fx_graph.get(node_name).get('real_siblings')
            if node_real_siblings is not None:
                is_node_real_siblings_needs_C_permutation = False
                for real_sibling_item in node_real_siblings:
                    if fx_graph.get(real_sibling_item).get('permutation_type') in ['C', 'KC']:
                        is_node_real_siblings_needs_C_permutation = True
                if is_node_real_siblings_needs_C_permutation == True:
                    print("[init_permutation_flag] node_name: \'{:}\', one of its real siblings need do offline permutation in C dim.".format(node_name))
                    node_original_permutation_type = fx_graph.get(node_name).get('permutation_type')
                    if node_original_permutation_type in ['C', 'KC']:
                        print("[init_permutation_flag] node_name: \'{:}\', its original permutation: \'{:}\' already includes C dim, no need to do No.1 post-processing change.".format(node_name, node_original_permutation_type))
                    elif node_original_permutation_type == 'None':
                        fx_graph[node_name]['permutation_type'] = 'C'
                        print("[init_permutation_flag] node_name: \'{:}\', change its original permutation: \'{:}\' to new permutation: 'C'.".format(node_name, node_original_permutation_type))
                        node_change_permutation_due_to_siblings.append(node_name)
                    elif node_original_permutation_type == 'K':
                        fx_graph[node_name]['permutation_type'] = 'KC'
                        print("[init_permutation_flag] node_name: \'{:}\', change its original permutation: \'{:}\' to new permutation: 'KC'.".format(node_name, node_original_permutation_type))
                        node_change_permutation_due_to_siblings.append(node_name)
        # Post-processing Step No.2:
        print("\n[init_permutation_flag] Post-processing Step No.2.")
        for node_name in fx_graph.keys():
            node_real_children = fx_graph.get(node_name).get('real_children')
            node_module_type = fx_graph.get(node_name).get('module_type')
            if (node_real_children is not None) and (node_module_type in ['torch.nn.modules.conv.Conv2d', 'torch.nn.modules.linear.Linear', 'torch.nn.modules.batchnorm.BatchNorm2d']):
                is_node_real_children_has_node_change_permutation = False
                for real_child_item in node_real_children:
                    if real_child_item in node_change_permutation_due_to_siblings:
                        is_node_real_children_has_node_change_permutation = True
                if is_node_real_children_has_node_change_permutation == True:
                    print("[init_permutation_flag] node_name: \'{:}\', one of its real children has changed permutation due to its siblings.".format(node_name))
                    node_original_permutation_type = fx_graph.get(node_name).get('permutation_type')
                    if node_original_permutation_type in ['K', 'KC']:
                        print("[init_permutation_flag] node_name: \'{:}\', its original permutation: \'{:}\' already includes K dim, no need to do No.2 post-processing change.".format(node_name, node_original_permutation_type))
                    elif node_original_permutation_type == 'None':
                        fx_graph[node_name]['permutation_type'] = 'K'
                        print("[init_permutation_flag] node_name: \'{:}\', change its original permutation: \'{:}\' to new permutation: 'K'.".format(node_name, node_original_permutation_type))
                    elif node_original_permutation_type == 'C':
                        fx_graph[node_name]['permutation_type'] = 'KC'
                        print("[init_permutation_flag] node_name: \'{:}\', change its original permutation: \'{:}\' to new permutation: 'KC'.".format(node_name, node_original_permutation_type))

        if cls.__save_permutation_graph:
            cls.save_graph_to_json(fx_graph, save_dumped_graph_path_with_name=os.path.join(cls.__permutation_output_dir, './model_graph_init_permutation_flag.json'))    # save the intermediate graph as JSON file for debugging
        return fx_graph

    @classmethod
    def extract_all_unique_siblings(cls, fx_graph):
        """This function is used to extrat all unique siblings for the whole network graph built with Torch.FX."""
        print("\n[extract_all_unique_siblings] Extract all unique siblings for the whole network graph built with Torch.FX")
        all_unique_siblings_name = []
        all_unique_siblings_module_type = []
        for node_name in fx_graph.keys():
            fx_graph[node_name]['node_type'] = 'network_node'    # use the 'node_type' to divide the real nodes apart from the auxiliary info node, like 'unique_siblings' node
            node_module_type = fx_graph.get(node_name).get('module_type')
            node_real_siblings = fx_graph.get(node_name).get('real_siblings')
            node_real_siblings_module_type = fx_graph.get(node_name).get('real_siblings_module_type')
            if node_real_siblings == []:
                print("[extract_all_unique_siblings] node_name: \'{:}\', node module type: \'{:}\', has no real siblings.".format(node_name, node_module_type))
                # for the Conv/FC layers without real_siblings, then we should insert itself as an unique_siblings
                if node_module_type in ['torch.nn.modules.conv.Conv2d', 'torch.nn.modules.linear.Linear']:
                    # direct insert will change the real_siblings info for the node in the fx_graph
                    node_real_siblings_with_node_itself = node_real_siblings.copy()
                    node_real_siblings_with_node_itself.insert(0, node_name)
                    node_real_siblings_module_type_with_node_itself = node_real_siblings_module_type.copy()
                    node_real_siblings_module_type_with_node_itself.insert(0, node_module_type)
                    all_unique_siblings_name.append(node_real_siblings_with_node_itself)
                    all_unique_siblings_module_type.append(node_real_siblings_module_type_with_node_itself)
            else:
                print("[extract_all_unique_siblings] node_name: \'{:}\', node module type: \'{:}\', has {:} real siblings: \'{:}\'.".format(node_name, node_module_type, len(node_real_siblings), node_real_siblings))
                # for the two duplicated siblings lists, the node names included should be the same.
                # If the node name is already included in one of the unique_siblings_name list, which means the real_siblings of this node is duplicated with the unique_siblings_name list.
                # Otherwise, we should insert the [real_siblings + node_name] as a new unique_siblings_name list.
                has_include_siblings = False
                for unique_siblings_item in all_unique_siblings_name:
                    if node_name in unique_siblings_item:
                        has_include_siblings = True
                if has_include_siblings == False:
                    # direct insert will change the real_siblings info for the node in the fx_graph
                    node_real_siblings_with_node_itself = node_real_siblings.copy()
                    node_real_siblings_with_node_itself.insert(0, node_name)
                    node_real_siblings_module_type_with_node_itself = node_real_siblings_module_type.copy()
                    node_real_siblings_module_type_with_node_itself.insert(0, node_module_type)
                    all_unique_siblings_name.append(node_real_siblings_with_node_itself)
                    all_unique_siblings_module_type.append(node_real_siblings_module_type_with_node_itself)

        fx_graph['unique_siblings'] = {}
        fx_graph['unique_siblings']['name'] = all_unique_siblings_name
        fx_graph['unique_siblings']['module_type'] = all_unique_siblings_module_type
        fx_graph['unique_siblings']['node_type'] = 'auxiliary_info_node'

        if cls.__save_permutation_graph:
            cls.save_graph_to_json(fx_graph, save_dumped_graph_path_with_name=os.path.join(cls.__permutation_output_dir, './model_graph_extract_all_unique_siblings.json'))    # save the intermediate graph as JSON file for debugging
        return fx_graph

    @classmethod
    def find_real_siblings(cls, fx_graph):
        """This function is used to find all siblings for each node according to the whole network graph built with Torch.FX.
        we need to find siblings recursively, because siblings may have siblings via other parents we don't know about.
        """
        print("\n[find_real_siblings] Find all siblings for each node according to the whole network graph built with Torch.FX")
        for node_name in fx_graph.keys():
            node_real_siblings_name = []
            node_real_siblings_module_type = []
            node_real_parents = fx_graph.get(node_name).get('real_parents')
            node_module_type = fx_graph.get(node_name).get('module_type')
            if node_module_type not in ['torch.nn.modules.conv.Conv2d', 'torch.nn.modules.linear.Linear']:
                print("[find_real_siblings] node_name: \'{:}\', node module type: \'{:}\', has no real siblings.".format(node_name, node_module_type))
            else:
                print("[find_real_siblings] node_name: \'{:}\', node module type: \'{:}\', may have real siblings.".format(node_name, node_module_type))
                # sibling means the nodes share the same real parent
                for real_parent_item in node_real_parents:
                    for real_child_item in fx_graph.get(real_parent_item).get('real_children'):
                        if real_child_item != node_name:
                            sibling_module_type = fx_graph.get(real_child_item).get('module_type')
                            print("[find_real_siblings] node_name: \'{:}\', has one real sibling: \'{:}\', its real sibling module type: \'{:}\'.".format(node_name, real_child_item, sibling_module_type))
                            node_real_siblings_name.append(real_child_item)
                            node_real_siblings_module_type.append(sibling_module_type)

            # remove the duplicated real siblings
            exclusive_node_real_siblings_name = []
            exclusive_node_real_siblings_module_type = []
            item_index = 0
            duplicated_real_siblings = 0
            for item in node_real_siblings_name:
                if item not in exclusive_node_real_siblings_name:
                    exclusive_node_real_siblings_name.append(item)
                    exclusive_node_real_siblings_module_type.append(node_real_siblings_module_type[item_index])
                else:
                    duplicated_real_siblings = duplicated_real_siblings + 1
                item_index = item_index + 1
            if duplicated_real_siblings > 0:
                print("[find_real_siblings] node_name: \'{:}\', remove {:} duplicated real siblings.".format(node_name, duplicated_real_siblings))
            fx_graph[node_name]['real_siblings'] = exclusive_node_real_siblings_name
            fx_graph[node_name]['real_siblings_module_type'] = exclusive_node_real_siblings_module_type

        if cls.__save_permutation_graph:
            cls.save_graph_to_json(fx_graph, save_dumped_graph_path_with_name=os.path.join(cls.__permutation_output_dir, './model_graph_find_real_siblings.json'))    # save the intermediate graph as JSON file for debugging
        return fx_graph

    @classmethod
    def recursive_find_real_children(cls, node_name, fx_graph):
        """This function is used to recursively find the real children for each node according to the whole network graph built with Torch.FX.
        Used as the sub-function of find_real_children.
        """
        node_real_children_name = []
        node_real_children_module_type = []
        if node_name in fx_graph.keys():    # can be deleted, because node_name is already in the 'children' item in one node of the fx_graph
            node_children = fx_graph.get(node_name).get('children')
            node_module_type = fx_graph.get(node_name).get('module_type')
            has_visit_children_num = 0
            has_real_children_num = 0
            sub_node_need_recursive_search = []
            while has_visit_children_num < len(node_children):
                for child_name in node_children:
                    if child_name != 'output':    # 'output' node has no 'module_type'
                        child_module_type = fx_graph.get(child_name).get('module_type')
                        if child_module_type in ['torch.nn.modules.conv.Conv2d', 'torch.nn.modules.linear.Linear']:
                            print("[recursive_find_real_children] node_name: \'{:}\', has one real child: \'{:}\', its real child module type: \'{:}\'.".format(node_name, child_name, child_module_type))
                            node_real_children_name.append(child_name)
                            node_real_children_module_type.append(child_module_type)
                            has_real_children_num = has_real_children_num + 1
                        else:
                            print("[recursive_find_real_children] node_name: \'{:}\', its child: \'{:}\' with module type: \'{:}\', needs recursive search.".format(node_name, child_name, child_module_type))
                            sub_node_need_recursive_search.append(child_name)
                    else:
                        print("[recursive_find_real_children] node_name: \'{:}\', its child: \'{:}\' with no module type, is not its real child.".format(node_name, child_name))
                    has_visit_children_num = has_visit_children_num + 1
            if len(sub_node_need_recursive_search) > 0:
                for sub_node in sub_node_need_recursive_search:
                    if fx_graph.get(sub_node).get('real_children') == []:
                        sub_node_real_children_name, sub_node_real_children_module_type = cls.recursive_find_real_children(sub_node, fx_graph)
                    else:
                        # if the sub_node already find the 'real_children', no need to do recursive search
                        sub_node_real_children_name = fx_graph.get(sub_node).get('real_children')
                        sub_node_real_children_module_type = fx_graph.get(sub_node).get('real_children_module_type')
                    node_real_children_name.extend(sub_node_real_children_name)
                    node_real_children_module_type.extend(sub_node_real_children_module_type)
        return node_real_children_name, node_real_children_module_type

    @classmethod
    def find_real_children(cls, fx_graph):
        """This function is used to find the real children for each node according to the whole network graph built with Torch.FX.
        For example:
        The real children of Conv is the subsequent Conv/FC.
        The real children of BN or other no-need-permutataion layers is the subsequent Conv/FC.
        """
        print("\n[find_real_children] Find the real children for each node according to the whole network graph built with Torch.FX")
        from sys import version_info
        if version_info.major == 3 and version_info.minor >= 8:
            reversible_fx_graph_keys = fx_graph.keys()
        else:    # 'dict_keys' object is not reversible in previous of Python 3.8
            reversible_fx_graph_keys = list(fx_graph.keys())
        for node_name in reversed(reversible_fx_graph_keys):    # as the optimization, we need to find the real children from back to front, to use the already saved 'real_children'
            node_real_children_name = []
            node_real_children_module_type = []
            node_children = fx_graph.get(node_name).get('children')
            node_module_type = fx_graph.get(node_name).get('module_type')
            if node_module_type not in ['torch.nn.modules.conv.Conv2d', 'torch.nn.modules.linear.Linear']:
                print("\n[find_real_children] node_name: \'{:}\', node module type: \'{:}\', children num: {:}, recursive to find real children.".format(node_name, node_module_type, len(node_children)))
                node_real_children_name, node_real_children_module_type = cls.recursive_find_real_children(node_name, fx_graph)
            else:    # Quick method, but cannot get the real children for no-need-permutataion layers like BN
                print("\n[find_real_children] node_name: \'{:}\', node module type: \'{:}\', children num: {:}, can directly find real children.".format(node_name, node_module_type, len(node_children)))
                # if the node is in the 'real_parents' list of the other node, then the other node is the real children for this node
                for other_node_name in fx_graph.keys():
                    if (other_node_name != node_name) and (node_name in fx_graph.get(other_node_name).get('real_parents')):
                        child_module_type = fx_graph.get(other_node_name).get('module_type')
                        if child_module_type in ['torch.nn.modules.conv.Conv2d', 'torch.nn.modules.linear.Linear']:
                            print("[find_real_children] node_name: \'{:}\', has one real child: \'{:}\', its real child module type: \'{:}\'.".format(node_name, other_node_name, child_module_type))
                            node_real_children_name.append(other_node_name)
                            node_real_children_module_type.append(child_module_type)

            # remove the duplicated real children
            exclusive_node_real_children_name = []
            exclusive_node_real_children_module_type = []
            item_index = 0
            duplicated_real_children = 0
            for item in node_real_children_name:
                if item not in exclusive_node_real_children_name:
                    exclusive_node_real_children_name.append(item)
                    exclusive_node_real_children_module_type.append(node_real_children_module_type[item_index])
                else:
                    duplicated_real_children = duplicated_real_children + 1
                item_index = item_index + 1
            if duplicated_real_children > 0:
                print("[find_real_children] node_name: \'{:}\', remove {:} duplicated real children.".format(node_name, duplicated_real_children))
            fx_graph[node_name]['real_children'] = exclusive_node_real_children_name
            fx_graph[node_name]['real_children_module_type'] = exclusive_node_real_children_module_type

        if cls.__save_permutation_graph:
            cls.save_graph_to_json(fx_graph, save_dumped_graph_path_with_name=os.path.join(cls.__permutation_output_dir, './model_graph_find_real_children.json'))    # save the intermediate graph as JSON file for debugging
        return fx_graph

    @classmethod
    def find_real_parents(cls, fx_graph):
        """This function is used to find the real parents for each node according to the whole network graph built with Torch.FX.
        For example:
        The real parent of BN   is the previous Conv/FC.
        The real parent of Conv is the previous Conv/FC.
        """
        print("\n[find_real_parents] Find the real parents for each node according to the whole network graph built with Torch.FX")
        for node_name in fx_graph.keys():
            node_real_parents_name = []
            node_real_parents_module_type = []
            node_parents = fx_graph.get(node_name).get('parents')
            print("[find_real_parents] node_name: \'{:}\', parents num: {:}".format(node_name, len(node_parents)))

            has_visit_parent_num = 0
            while has_visit_parent_num < len(node_parents):
                for parent_name in node_parents:
                    if fx_graph.__contains__(parent_name):
                        parent_module_type = fx_graph.get(parent_name).get('module_type')
                        if parent_module_type in ['torch.nn.modules.conv.Conv2d', 'torch.nn.modules.linear.Linear']:
                            print("[find_real_parents] node_name: \'{:}\', has one real parent: \'{:}\', its real parent module type: \'{:}\'.".format(node_name, parent_name, parent_module_type))
                            node_real_parents_name.append(parent_name)
                            node_real_parents_module_type.append(parent_module_type)
                        else:
                            print("[find_real_parents] node_name: \'{:}\', has one/several real parent(s): \'{:}\', its real parent module type: \'{:}\'.".format(node_name, fx_graph[parent_name]['real_parents'], fx_graph[parent_name]['real_parents_module_type']))
                            for real_parent_item in fx_graph[parent_name]['real_parents']:
                                node_real_parents_name.append(real_parent_item)
                            for real_parent_module_type_item in fx_graph[parent_name]['real_parents_module_type']:
                                node_real_parents_module_type.append(real_parent_module_type_item)
                    else:
                        print("[find_real_parents] node_name: \'{:}\', has no real parent because this is the first node.".format(node_name))
                    has_visit_parent_num = has_visit_parent_num + 1

            # remove the duplicated real parents
            exclusive_node_real_parents_name = []
            exclusive_node_real_parents_module_type = []
            exclusive_node_real_parents_groups_param = []
            item_index = 0
            duplicated_real_parents = 0
            for item in node_real_parents_name:
                if item not in exclusive_node_real_parents_name:
                    exclusive_node_real_parents_name.append(item)
                    exclusive_node_real_parents_module_type.append(node_real_parents_module_type[item_index])
                    exclusive_node_real_parents_groups_param.append(fx_graph.get(item).get('groups_param'))
                else:
                    duplicated_real_parents = duplicated_real_parents + 1
                item_index = item_index + 1
            if duplicated_real_parents > 0:
                print("[find_real_parents] node_name: \'{:}\', remove {:} duplicated real parents.".format(node_name, duplicated_real_parents))
            fx_graph[node_name]['real_parents'] = exclusive_node_real_parents_name
            fx_graph[node_name]['real_parents_module_type'] = exclusive_node_real_parents_module_type
            fx_graph[node_name]['real_parents_groups_param'] = exclusive_node_real_parents_groups_param

        if cls.__save_permutation_graph:
            cls.save_graph_to_json(fx_graph, save_dumped_graph_path_with_name=os.path.join(cls.__permutation_output_dir, './model_graph_find_real_parent.json'))    # save the intermediate graph as JSON file for debugging
        return fx_graph

    @classmethod
    def build_fx_graph(cls, model, dump_fx_graph=False, save_dumped_fx_graph='./model_fx_graph.json'):
        """This function is used to build the whole network graph with Torch.FX features."""
        success = True
        torch_version = str(torch.__version__)
        torch_version_major = int(torch_version.split('.')[0])
        torch_version_minor = int(torch_version.split('.')[1])
        try:
            torch_version_minimum = int(torch_version.split('.')[2])
        except ValueError:    # support the none standard version
            torch_version_minimum = torch_version.split('.')[2]
        print("[build_fx_graph] The torch version is: {}, version major is: {}, version minor is: {}, version minimum is: {}".format(torch_version, torch_version_major, torch_version_minor, torch_version_minimum))
        if torch_version_major >= 1 and torch_version_minor >= 8:
            print("[build_fx_graph] The Torch.FX is supported.")
        else:    # Torch.FX is introduced in torch 1.8.0
            print("[build_fx_graph] The Torch.FX is not supported. So cannot build the Torch.FX graph.")
            success = False
            network_fx_graph = {}
            return network_fx_graph, success

        print("\n[build_fx_graph] Print the model structure with pure PyTorch function")
        print(model)

        print("\n[build_fx_graph] Build the module name and type dictionary")
        module_name_type_dict = {}
        module_name_group_conv_dict = {}
        for name, mod in model.named_modules():
            print("[build_fx_graph] module_name: {}, module type: {}".format(name, type(mod)))
            module_name_type_dict[name] = str(type(mod)).split("\'")[1]
            try:
                print("[build_fx_graph] this module has \'group\' param with value: {}".format(mod.groups))
                module_name_group_conv_dict[name] = str(mod.groups)
            except:
                module_name_group_conv_dict[name] = 'None'
                continue

        graph_module = cls.print_raw_fx_graph(model, print_tabular=True)

        # keep track of children and parents for each layer (could be call_module or call_function)
        print("\n[build_fx_graph] Print the children and parents relationship for each layer")
        network_fx_graph = {}
        for node in graph_module.graph.nodes:
            if node.op == 'placeholder':
                print("[build_fx_graph] This is the \'input\' node: {:}".format(node.target))
                continue
            elif node.op == 'get_attr':
                print("[build_fx_graph] This is the \'get_attr\' node: {:}".format(node.target))
                continue
            elif node.op == 'call_function':    # e.g. 'adaptive.avg.pool2d', 'add', 'cat', 'flatten', 'floordiv', 'getattr', 'getitem', 'hardsigmoid', 'mean', 'mul', 'relu', 'transpose'
                node_parent, node_children = get_node_parent_children(node)
                converted_node_name=convert_fx_node_name(node.name)
                print("[build_fx_graph] This is the \'call_function\' node: {:}, its parent list: {:}, its children list: {:}".format(converted_node_name, node_parent, node_children))
                network_fx_graph[converted_node_name] = {}
                network_fx_graph[converted_node_name]['parents'] = node_parent
                network_fx_graph[converted_node_name]['children'] = node_children
                network_fx_graph[converted_node_name]['fx_op'] = 'call_function'
            elif node.op == 'call_method':    # e.g. 'chunk', 'contiguous', 'mean', 'size', 'unsqueeze', 'view'
                node_parent, node_children = get_node_parent_children(node)
                converted_node_name=convert_fx_node_name(node.name)
                print("[build_fx_graph] This is the \'call_method\' node: {:}, its parent list: {:}, its children list: {:}".format(converted_node_name, node_parent, node_children))
                network_fx_graph[converted_node_name] = {}
                network_fx_graph[converted_node_name]['parents'] = node_parent
                network_fx_graph[converted_node_name]['children'] = node_children
                network_fx_graph[converted_node_name]['fx_op'] = 'call_method'
                continue
            elif node.op == 'call_module':
                node_parent, node_children = get_node_parent_children(node)
                converted_node_name=convert_fx_node_name(node.name)
                # check whether the converted_node_name is same as node.target, especially for ReLU case
                if converted_node_name != node.target:
                    print("[build_fx_graph][warning] The target name from Torch.FX is \'{:}\', the manually converted node name is \'{:}\', not the same one, choose the converted node name".format(node.target, converted_node_name))
                # assume the modules share the same target name have the same type, because converted_node_name may not be obtained by model.named_modules(), like some ReLU (defined in forward function)
                node_type = module_name_type_dict[node.target]
                print("[build_fx_graph] This is the \'call_module\' node: {:}, its parent list: {:}, its children list: {:}, its type: {:}".format(converted_node_name, node_parent, node_children, node_type))
                network_fx_graph[converted_node_name] = {}
                network_fx_graph[converted_node_name]['parents'] = node_parent
                network_fx_graph[converted_node_name]['children'] = node_children
                network_fx_graph[converted_node_name]['fx_op'] = 'call_module'
                network_fx_graph[converted_node_name]['module_type'] = node_type
                network_fx_graph[converted_node_name]['groups_param'] = module_name_group_conv_dict[node.target]
            elif node.op == 'output':
                print("[build_fx_graph] This is the \'output\' node: {:}".format(node.target))
                continue

        if dump_fx_graph:
            print("\n[build_fx_graph] Dump the overall dict for children and parents relationship into JSON file")
            cls.save_graph_to_json(network_fx_graph, save_dumped_graph_path_with_name=save_dumped_fx_graph)

        return network_fx_graph, success

    @classmethod
    def print_raw_fx_graph(cls, model, print_tabular=False, generate_python_code=False):
        """This function is used to print the intermediate representation (IR) - Graph representation with Torch.FX features."""
        from torch.fx import symbolic_trace
        # Symbolic tracing frontend - captures the semantics of the module
        try:
            symbolic_traced : torch.fx.GraphModule = symbolic_trace(model)
        except:
            if not torch.distributed.is_initialized() or torch.distributed.get_rank() == 0:
                print("\n[print_raw_fx_graph] Meet the fatal fault when trying to symbolic trace the model with Torch.FX")
                raise
            exit(0)

        # High-level intermediate representation (IR) - Graph representation
        print("\n[print_raw_fx_graph] Print the intermediate representation (IR) with Torch.FX")
        print(symbolic_traced.graph)

        if print_tabular:
            print("\n[print_raw_fx_graph] Print the intermediate representation (IR) with Torch.FX in a table format")
            try:
                symbolic_traced.graph.print_tabular()
            except AttributeError:    # to avoid the AttributeError: 'Graph' object has no attribute 'print_tabular'
                print("[print_raw_fx_graph][Warning] \'print_tabular\' function is not supported in current Torch version. Skip!")

        # Code generation - valid Python code
        if generate_python_code:
            print("\n[print_raw_fx_graph] Create valid Python code matching the IR/Graph's semantics with Torch.FX")
            print(symbolic_traced.code)

        return symbolic_traced

    @classmethod
    def save_graph_to_json(cls, graph, save_dumped_graph_path_with_name='./model_fx_graph.json'):
        """This function is used to same the graph into JSON file."""
        # use dumps to transfer the dict to JSON string
        json_graph_str = json.dumps(graph)
        with open(save_dumped_graph_path_with_name, 'w', encoding='utf-8') as dumped_graph_file:
            dumped_graph_file.write(json_graph_str)  # write the transferred JSON string into JSON file
